--[[To do:
Swap equip key
Drag equipped items to bank
Multiple slot counts
Close all / expand all
Remove all categories
Data transfer to other players
Items with charges
Equip Slot indicator
Alternate color indicator for special slot counts
Find solution for "0" item buildup - have to set a lock?
Shift key for forced arrangement
]]

BaudManifestDebug = false;

--if extra backgrounds are added, they must be formatted to fit WoW's standards and the client must be restarted
local Backgrounds = {
  {File = false, Size = 8},
  {File = "Asphalt", Size = 128},
  {File = "Bricks", Size = 128},
  {File = "Bumpy", Size = 64},
  {File = "CrackedPaint", Size = 128},
  {File = "DiamondPlate", Size = 128},
  {File = "Grain", Size = 128},
  {File = "Hatch", Size = 128},
  {File = "Paper", Size = 128},
};

local DragItem, DragPending, VirtualDrag, BankOpen, TradeOpen, VendorOpen, MailOpen;
local SplitHoldItem, SplitHoldExpire, SplitPending, SplitTimeout, IgnoreHooks, CursorItem;
local DisableCombine;
local CursorHover, PlayerName, SelectedChar;
local RealmData, PlayerData, AllCharData, AllCharHash, OtherCharMoney, BagTypes;
local PlayerHash = {};
local DisplayNames = {"Inventory","Bank Box"};
--List will refresh on first OnUpdate
local QueuedUpdate = {true, true};
local QueuedUpdateFrame = CreateFrame("Frame");
local ClearCursorFrame = CreateFrame("Frame");
local CursorHoverFrame = CreateFrame("Frame");
CursorHoverFrame:Hide();
local ItemStringMatch = "(item:(%d+):[%d:]+:(%-?%d+)):(%-?%d+)|";
local Tooltip;
local _; --Space filler variable.  Declared anyways to avoid writing to globals.
local HookData = {};
local MsgPrefix = "|cffffff00Baud Manifest:|r ";

local FilterDelay = 0.5;
local FadeTime = 0.2;
local IndentSize = 10;

local CheckButtons = {
  {Text="Show Fading",SavedVar="Fading",Default=true,
    TooltipText="The manifest will fade in when opened, and fade out when closed."},
  {Text="Auto Combine Stacks",SavedVar="Combine",Default=true,
    TooltipText="Automatically combines incomplete item stacks."},
  {Text="Auto Fill Special Bags",SavedVar="FillSpecials",Default=true,
    TooltipText="Automatically moves items from normal bags into specialty ones (quiver, etc.).  This helps keep your all-purpose bag slots free for use."},
  {Text="Auto Show Inventory",SavedVar="AutoShowInv",Default=true,
    TooltipText="Automatically shows your inventory at the bank, mailbox, vendor, auction house, or while trading."},
  {Text="Stop Blizzard's Auto Show",SavedVar="HideBlizInv",Default=true,
    TooltipText="Stops Blizzard's backpack from automatically opening at the mail and vendor."},
  {Text="Auto Show Bank",SavedVar="ShowBank",Default=true,
    TooltipText="Shows your bank box when it is opened.  The bank box can also be opened with a hotkey, or from the inventory's menu."},
  {Text="Hide Blizzard's Bank",SavedVar="HideBlizBank",Default=true, Depend="ShowBank",
    TooltipText="Prevents Blizzard's bank box from showing when you open your bank box."},
  {Text="Replace Blizzard's Bags",SavedVar="ReplaceBlizBags",Default=true,
    TooltipText="Opening Blizzard's bags will open Baud Manifest instead."},
  {Text="Hide Equipped",SavedVar="HideEquipped",Default=true,
    TooltipText="Hides items in the manifest when they are equipped.  Equipped items will always be shown when the list is filtered, sorted, or viewed from another character."},
  {Text="Item Name First",SavedVar="ItemNameFirst",Default=false,
    TooltipText="Displays the item's name before its quantity."},
  {Text="Show Arrows",SavedVar="ShowArrows",Default=true,
    TooltipText="Shows arrows at the top or bottom of the manifest if not all items can be displayed at once."},
};

--[[local SliderBars = {
  {Text="Scale - %d%%",Low="50%",High="200%",SavedVar="Scale", Default=100,
    TooltipText = "Scale of the manifest."}
};]]

--UIPanelWindows["BaudManifestCharacters"] = {area="center", pushable = 0, whileDead = 1};
--This list controls which variables in the item list get saved.  Extra data is discarded on log out to save memory.
local SaveItemVars = {
  ItemString = true,
  SuffixFactor = true,
  Category = true,
  Count = true,
  Slots = true,
  Equipped = true,
};

local Sorts = {
  {Name = "No Sorting"},
  {Name = "Alphabetical", Func = function(ListItem)
    return select(1,GetItemInfo(ListItem.ItemString));
  end},
  {Name = "Rarity", Func = function(ListItem)
    return select(3,GetItemInfo(ListItem.ItemString));
  end},
  {Name = "Item Type", Func = function(ListItem)
    local Name, _, _, _, _, Type, SubType, _, EquipLoc = GetItemInfo(ListItem.ItemString);
    if(EquipLoc=="INVTYPE_ROBE")then
      EquipLoc = "INVTYPE_CHEST";
    elseif(EquipLoc=="INVTYPE_THROWN")then
      EquipLoc = "INVTYPE_RANGED";
    elseif(EquipLoc=="INVTYPE_HOLDABLE")or(EquipLoc=="INVTYPE_SHIELD")then
      EquipLoc = "INVTYPE_WEAPONOFFHAND2";
    end
    return Type..EquipLoc..SubType;
  end},
  {Name = "Slots Used", Func = function(ListItem)
    return -(not ListItem.Count and 0 or ListItem.Slots or 1);
  end},
};

--Traveller's Backpack, Netherweave Bag, Imbued Netherweave Bag, Mooncloth Bag, Primal Mooncloth Bag
local StandardBags = {4500, 21841, 21843, 14155, 21876};

local InvBagsOffset = ContainerIDToInventoryID(1) - 1;
local BankBagsOffset = ContainerIDToInventoryID(5) - 1;
local strmatch = string.match;

local function DebugMsg(Msg)
  if BaudManifestDebug then
    DEFAULT_CHAT_FRAME:AddMessage(Msg);
  end
end


StaticPopupDialogs["BAUDMANIFEST_DELETECATEGORY"] = {
  text = "Are you sure you want to delete this category?",
  button1 = TEXT(YES),
  button2 = TEXT(NO),
  OnAccept = function()
    BaudManifestDeleteDraggedCategory();
  end,
  OnCancel = function()
    DebugMsg("Cancel delete category.");
    BaudManifestClearVirtualDrag();
  end,
  timeout = 0,
  exclusive = 1,
  hideOnEscape = 1
};


StaticPopupDialogs["BAUDMANIFEST_CATEGORIZE"] = {
  text = "This action will put uncategorized items into categories based on their type.  Do you wish to proceed?",
  button1 = TEXT(YES),
  button2 = TEXT(NO),
  OnAccept = function(Display)
    BaudManifestCategorize(Display);
  end,
  timeout = 0,
  exclusive = 1,
  hideOnEscape = 1
};


local function IsDisplayAccessable(Display)
  return (Display:GetID()==1)or BankOpen and(Display:GetID()==2);
end


function BaudManifestToggleDisplay(Display)
  Display = getglobal("BaudManifestDisplay"..Display);
  if Display:IsShown()then
    Display:Hide();
  else
    Display:Show();
  end
end


function BaudManifestToggleChars()
  if BaudManifestCharacters:IsShown()then
    BaudManifestCharacters:Hide()
  else
    BaudManifestCharacters:ClearAllPoints();
    BaudManifestCharacters:SetPoint("CENTER");
    BaudManifestCharacters:Show();
  end
end


function BaudManifest_OnLoad(self)
  BaudManifestVersionText:SetText("Version "..GetAddOnMetadata("BaudManifest","Version"));

  SlashCmdList["BaudManifest"] = function(Msg)
    if(strlower(Msg)=="options")then
      BaudManifestOptions:Show();
    else
      BaudManifestToggleDisplay(1);
    end
  end
  Tooltip = BaudManifestTooltip;
  SLASH_BaudManifest1 = "/baudmanifest";
  SLASH_BaudManifest2 = "/baudman";
  SLASH_BaudManifest3 = "/bm";
  SLASH_BaudManifestMenuName = "Baud Manifest";
  SLASH_BaudManifestMenuArgs = "Options";
  BINDING_HEADER_BaudManifest = "Baud Manifest";
  BINDING_NAME_BaudManifest = "Toggle Inventory";
  BINDING_NAME_BaudManifestToggleBank = "Toggle Bank";
  BINDING_NAME_BaudManifestToggleChars = "Toggle Characters";
  self:RegisterEvent("VARIABLES_LOADED");
  self:RegisterEvent("BAG_UPDATE");
  self:RegisterEvent("UNIT_INVENTORY_CHANGED");
  self:RegisterEvent("BAG_CLOSED");
  self:RegisterEvent("BAG_UPDATE_COOLDOWN");
  self:RegisterEvent("PLAYERBANKSLOTS_CHANGED");
  self:RegisterEvent("PLAYERBANKBAGSLOTS_CHANGED");
  self:RegisterEvent("BANKFRAME_OPENED");
  self:RegisterEvent("BANKFRAME_CLOSED");
  self:RegisterEvent("TRADE_SHOW");
  self:RegisterEvent("TRADE_CLOSED");
  self:RegisterEvent("TRADE_PLAYER_ITEM_CHANGED");
  self:RegisterEvent("MAIL_SHOW");
  self:RegisterEvent("MAIL_CLOSED");
  self:RegisterEvent("MERCHANT_SHOW");
  self:RegisterEvent("MERCHANT_CLOSED");
  self:RegisterEvent("AUCTION_HOUSE_SHOW");
  self:RegisterEvent("AUCTION_HOUSE_CLOSED");
  self:RegisterEvent("NEW_AUCTION_UPDATE");
  self:RegisterEvent("CURSOR_UPDATE");
  self:RegisterEvent("ITEM_LOCK_CHANGED");
  self:RegisterEvent("PLAYER_MONEY");
  self:RegisterEvent("PLAYER_LOGIN");
  self:RegisterEvent("PLAYER_LOGOUT");
  DEFAULT_CHAT_FRAME:AddMessage("Baud Manifest: AddOn loaded. Type /baudman to display your manifest.");
end


local function GetCharData(Name)
  if(type(RealmData[Name])~="table")then
    RealmData[Name] = {};
  end
  local CharData = RealmData[Name];
  for Index = 1, 2 do
    if(type(CharData[Index])~="table")then
      CharData[Index] = {};
    end

    if(type(CharData[Index][1])~="table")then
      CharData[Index][1] = {};
    end

    if(type(CharData[Index].BGColor)~="table")then
      CharData[Index].BGColor = {0.5,0.5,0.5,0.7};
    end
    local Valid;
    for Key, Value in ipairs(Backgrounds)do
      if(Value.File==CharData[Index].BGTexture)then
        Valid = true;
        break;
      end
    end
    if not Valid then
      CharData[Index].BGTexture = "Bumpy";
    end
  end
  return CharData;
end


local function SaveBankBagData()
  for Bag = 1, NUM_BANKBAGSLOTS do
    PlayerData.BankBag[Bag].Link = GetInventoryItemLink("player", Bag + BankBagsOffset);
    PlayerData.BankBag[Bag].Count = GetInventoryItemCount("player", Bag + BankBagsOffset);
  end
end


function BaudManifest_OnEvent(self, event)
  if(event=="VARIABLES_LOADED")then
    if(type(BaudManifestData)~="table")then
      BaudManifestData = {};
    end
    local RealmName = GetRealmName();
    if(type(BaudManifestData[RealmName])~="table")then
      BaudManifestData[RealmName] = {};
    end
    RealmData = BaudManifestData[RealmName];
    PlayerName = UnitName("player");
    --Copy from individual character saved variables for backwards compatibility
    if not RealmData[PlayerName]and BaudManifest_Cfg then
      RealmData[PlayerName] = BaudManifest_Cfg.Items;
    end
    PlayerData = GetCharData(PlayerName);
    if(type(PlayerData.Scale)~="number")then
      PlayerData.Scale = 100;
    end
    BaudManifestUpdateScales();

    local Hash;
    for Index = 1, 2 do
      local Display = getglobal("BaudManifestDisplay"..Index);
      Display.Data = PlayerData[Index];
      BaudManifestUpdateBGTexture(Display);
      getglobal(Display:GetName().."BagsButton"):SetChecked(Display.Data.ShowBags);
      if Display.Data.ShowBags then
        Display.BagsFrame:Show();
      else
        Display.BagsFrame:Hide();
      end
      Hash = {};
      PlayerHash[Index] = Hash;
      Display.Hash = Hash;
      for Category, Value in ipairs(PlayerData[Index])do
        for Key, Value in ipairs(Value)do
          if Value.ItemString then
            Hash[Value.ItemString] = Value;
          end
        end
      end
    end
    BaudManifestUpdateBlizBankBlock();
    if PlayerData.ReplaceBlizBags then
      BaudManifestHookBlizBags();
    end
    if not PlayerData.BankBag then
      PlayerData.BankBag = {};
    end
    local Texture, BagSlot;
    for Bag = 1, NUM_BANKBAGSLOTS do
      --Bag name length is restricted
      BagSlot = getglobal("BaudMBankBag"..Bag);
      Texture = nil;
      if not PlayerData.BankBag[Bag]then
        PlayerData.BankBag[Bag] = {};
      elseif PlayerData.BankBag[Bag].Link then
        Texture = select(10, GetItemInfo(PlayerData.BankBag[Bag].Link));
        SetItemButtonCount(BagSlot, PlayerData.BankBag[Bag].Count or 0);
      else
        Texture = select(2, GetInventorySlotInfo("Bag"..Bag));
      end
      SetItemButtonTexture(BagSlot,Texture);
    end

    local CharList = {};
    for Name, Value in pairs(RealmData)do
      if(Name~=PlayerName)and(Name~="All Characters")then
        tinsert(CharList,Name);
      end
    end
    table.sort(CharList,function(a,b) return(a < b);end);
    tinsert(CharList,1,"All Characters");
    local Entry, Texture, Coords, Text;
    for Key, Value in ipairs(CharList)do
      Entry = CreateFrame("Button","BaudManifestCharacterEntry"..Key,BaudManifestCharactersScrollBox,"BaudManifestCharacterEntryTemplate");
      Entry:SetPoint("TOP",0,6-16*Key);
      Text = getglobal(Entry:GetName().."Text");
      Text:SetText(Value);
      Entry.Name = Value;
      if(Key==1)then
        Text:SetTextColor(1,1,1);
        Text:SetPoint("LEFT",2,0);
      else
        Texture = getglobal(Entry:GetName().."Texture");
        Coords = CLASS_BUTTONS[strupper(RealmData[Value].Class or "")];
        if Coords then
      	  Texture:SetTexture("Interface\\WorldStateFrame\\Icons-Classes");
    	    Texture:SetTexCoord(unpack(Coords));
    	    Texture:Show();
        end
      end
    end

    if(type(PlayerData.PutQuality)~="table")then
      PlayerData.PutQuality = {};
    end
    for Quality = 0, 2 do
      if not PlayerData.PutQuality[Quality]then
        PlayerData.PutQuality[Quality] = 1;
      end
    end
    if not PlayerData.AutoSell then
      PlayerData.AutoSell = 1;
    end
    for Index, Value in ipairs(CheckButtons)do
      if(PlayerData[Value.SavedVar]==nil)then
        PlayerData[Value.SavedVar] = Value.Default;
      end
    end

    if(type(BaudManifestBagTypeData)~="table")then
      BagTypes = {};
      BaudManifestBagTypeData = BagTypes;
    else
      BagTypes = BaudManifestBagTypeData;
    end

  elseif(event=="PLAYER_LOGIN")then
    PlayerData.Class = UnitClass("player");
    PlayerData.Money = GetMoney();

    local Standard;
    for Key, Value in ipairs(StandardBags)do
      Standard = select(7, GetItemInfo(Value));
      if Standard then
        BagTypes.Standard = Standard;
        BagTypes[Standard] = nil;
        break;
      end
    end
    if not BagTypes.Standard then
      BagTypes.Standard = BaudManifestStandardBag;
    end
    BaudManifestBankPurchaseUpdate();

  elseif(event=="PLAYER_LOGOUT")then
    --Variable cleanup here
    for Display, Value in ipairs(PlayerData)do
      for Category, Value in ipairs(Value)do
        for ItemIndex, Value in ipairs(Value)do
          for Key, _ in pairs(Value)do
            if not SaveItemVars[Key]then
              Value[Key] = nil;
            end
          end
        end
      end
    end

  elseif(event=="PLAYER_MONEY")then
    PlayerData.Money = GetMoney();

  elseif(event=="BAG_UPDATE")or(event=="BAG_CLOSED")or(event=="PLAYERBANKSLOTS_CHANGED")then
    if(event=="PLAYERBANKSLOTS_CHANGED")then
      if(arg1 > NUM_BANKGENERIC_SLOTS)then
        BankFrameItemButton_Update(getglobal("BaudMBankBag"..(arg1-NUM_BANKGENERIC_SLOTS)));
        return;
      end
    end
    --Updating of items done even if the bag is closed to keep the ordering correct
    local Display;
    if(event=="PLAYERBANKSLOTS_CHANGED")or(arg1==-1)or(arg1 > 4)then
      Display = 2;
    else
      Display = 1;
    end
    QueuedUpdate[Display] = true;
    QueuedUpdateFrame:Show();
    if DragPending and(DragPending.Origin:GetID()==Display)then
      DragPending.Ready = true;
    end

  elseif(event=="UNIT_INVENTORY_CHANGED")then
    if(arg1~="player")then
      return;
    end
    QueuedUpdate[1] = true;
    QueuedUpdateFrame:Show();

  elseif(event=="ITEM_LOCK_CHANGED")then
    local Bag, Slot = arg1, arg2;
    if(Bag == BANK_CONTAINER)and(Slot > NUM_BANKGENERIC_SLOTS)then
      BankFrameItemButton_UpdateLocked(getglobal("BaudMBankBag"..(Slot-NUM_BANKGENERIC_SLOTS)));
      return;
    end
    --The list is updated to change the specific targeted items, but only if the list is open
    for Display = 1, 2 do
      if getglobal("BaudManifestDisplay"..Display):IsShown()then
        QueuedUpdate[Display] = true;
      end
    end
    QueuedUpdateFrame:Show();

  elseif(event=="BAG_UPDATE_COOLDOWN")then
    BaudManifestScrollBar_Update(BaudManifestDisplay1);

  elseif(event=="PLAYERBANKBAGSLOTS_CHANGED")then
    BaudManifestBankPurchaseUpdate();

  elseif(event=="BANKFRAME_OPENED")then
    if(event=="BANKFRAME_OPENED")then
      BankOpen = true;
      BaudManifestUpdateList(BaudManifestDisplay2);
      BaudManifestAutoShow(false, true);
      BaudManifestBankSlotPurchaseButton:Enable();
    end
    SaveBankBagData();
    for Index = 1, NUM_BANKBAGSLOTS do
      BankFrameItemButton_Update(getglobal("BaudMBankBag"..Index));
    end

  elseif(event=="BANKFRAME_CLOSED")then
    BankOpen = false;
    BaudManifestAutoShow(true, true);
    BaudManifestBankSlotPurchaseButton:Disable();

  elseif(event=="TRADE_SHOW")then
    TradeOpen = true;
    if not CursorHasItem()then
      BaudManifestPickupDragRemain(BaudManifestTradeRemain);
    end
    BaudManifestAutoShow();

  elseif(event=="TRADE_CLOSED")then
    TradeOpen = false;
    QueuedUpdate[1] = true;
    QueuedUpdateFrame:Show();
    BaudManifestAutoShow(true);

  --This is needed to detect if the item gets dropped on a player in the game world.  Problem is, it fires before the item is
  --Removed from the cursor.
  elseif(event=="TRADE_PLAYER_ITEM_CHANGED")then
    if TradeOpen and CursorHasItem()and DragItem and not IgnoreHooks then
      DragItem.Trading = true;
    end

  elseif(event=="MAIL_SHOW")then
    MailOpen = true;
    BaudManifestAutoShow();

  elseif(event=="MAIL_CLOSED")then
    MailOpen = false;
    BaudManifestAutoShow(true);

  elseif(event=="MERCHANT_SHOW")then
    VendorOpen = true;
    BaudManifestAutoSellItems();
    BaudManifestAutoShow();

  elseif(event=="MERCHANT_CLOSED")then
    VendorOpen = false;
    BaudManifestAutoShow(true);

  elseif(event=="AUCTION_HOUSE_SHOW")then
    BaudManifestAutoShow();

  elseif(event=="AUCTION_HOUSE_CLOSED")then
    BaudManifestAutoShow(true);

  elseif(event=="NEW_AUCTION_UPDATE")then
    --Prevents Baud Manifest from combining items split apart for auctioning, after the auction gets posted, and before the
    --item disappears from your pack.  This is harmless anyways, but it is unnecissary, so I eliminate it.
    DisableCombine = GetTime() + 1;

  elseif(event=="CURSOR_UPDATE")then
    if(arg1=="up")then
      BaudManifestClearFocus();
    end
    if(CursorHasItem()~=CursorItem)then
      CursorItem = CursorHasItem();
      if not CursorItem then
        ClearCursorFrame:Show();
        BaudManifestDragCount:Hide();
        if DragItem and DragItem.Trading then
          DragItem.Trading = nil;
          if not IgnoreHooks then
            BaudManifestPickupDragRemain(BaudManifestTradeRemain);
          end
        end

      elseif DragItem and not IgnoreHooks then
        DragItem = nil;
      end
      BaudManifestUpdateBlocked();
    end
  end
end


function BaudManifestAutoShow(Hide, Bank)
  local Display;
  for Index = 1, 2 do
    if(Index==1)and PlayerData.AutoShowInv or(Index==2)and Bank and PlayerData.ShowBank then
      Display = getglobal("BaudManifestDisplay"..Index);
      if Hide then
        if Display:IsShown()then
          if Display.AutoShown then
            Display:Hide();
            DebugMsg("Auto hiding display "..Index);
          else
            BaudManifestScrollBar_Update(Display);
          end
        end
      elseif not Display:IsShown()or Display.Closing then
        Display:Show();
        Display.Closing = false;
        Display.AutoShown = true;
        DebugMsg("Auto showing display "..Index);
      end
    end
  end
end


function BaudManifestCategoryEditBox_OnEditFocusLost(self)
  if self:IsShown()then
    self.Category.Name = self:GetText();
  end
  self:Hide();
  BaudManifestScrollBar_Update(self.Display);
end


function BaudManifestClearVirtualDrag()
  if not VirtualDrag then
    return;
  end
  DebugMsg("Clearing virtual drag.");
  PlaySound("INTERFACESOUND_CURSORDROPOBJECT");
  VirtualDrag = nil;
  BaudManifestVirtualDrag:Hide();
  BaudManifestUpdateBlocked();
end


function BaudManifestDeleteDraggedCategory()
  if not VirtualDrag or not VirtualDrag.Category then
    DebugMsg("No category to delete.");
    return;
  end
  DebugMsg("Deleting category #"..VirtualDrag.Category);

  local Replaced, ListIndex;
  for Category, ParentCat in ipairs(VirtualDrag.Origin.Data)do
    ListIndex = 1;
    while ParentCat[ListIndex]do
      if ParentCat[ListIndex].Category then
        if(ParentCat[ListIndex].Category==VirtualDrag.Category)then
          DebugMsg("Located category listing, removing and replacing it with its contents.");
          tremove(ParentCat, ListIndex);
          ListIndex = ListIndex - 1;
          --Now replace the deleted category with its contents
          if not Replaced then
            Replaced = true;
            for Key, Value in ipairs(VirtualDrag.Origin.Data[VirtualDrag.Category])do
              --Make sure the copied entries are skipped, or categories could be decremented twice
              ListIndex = ListIndex + 1;
              tinsert(ParentCat,ListIndex,Value);
            end
          end
        elseif(ParentCat[ListIndex].Category > VirtualDrag.Category)then
          ParentCat[ListIndex].Category = ParentCat[ListIndex].Category - 1;
        end
      end
      ListIndex = ListIndex + 1;
    end
  end
  if(VirtualDrag.Origin:GetID()==1)then
    for Quality, Value in pairs(PlayerData.PutQuality)do
      if(Value==VirtualDrag.Category)then
        PlayerData.PutQuality[Quality] = 1;
      elseif(Value > VirtualDrag.Category)then
        PlayerData.PutQuality[Quality] = Value - 1;
      end
    end
    if(PlayerData.AutoSell==VirtualDrag.Category)then
      PlayerData.AutoSell = 1;
    elseif(PlayerData.AutoSell > VirtualDrag.Category)then
      PlayerData.AutoSell = PlayerData.AutoSell - 1;
    end
  end

  DebugMsg("Deleting category variable.");
  tremove(VirtualDrag.Origin.Data,VirtualDrag.Category);
  BaudManifestUpdateDisplayList(VirtualDrag.Origin);
  BaudManifestClearVirtualDrag();
end


function BaudManifestCategoriesDisabled(Display)
  return (Display.Sort~=1)or(Display.Filter~=1)or strfind(Display.CustomFilter or "","%S");
end


function BaudManifestCategorize(Display)
  DebugMsg("Categorizing loose items");
  local Type, SubType, Category, New;
  local Index = 1;
  while(Display.Data[1][Index])do
    if Display.Data[1][Index].ItemString then
      _, _, _, _, _, Type, SubType = GetItemInfo(Display.Data[1][Index].ItemString);
      if not Type then
        Index = Index + 1;
      else
        Category, New = BaudManifestGetCategory(Display, Type, 1);
        if New then
          Index = Index + 1;
        end
        tinsert(Display.Data[Category],Display.Data[1][Index]);
        tremove(Display.Data[1],Index);
      end
    else
      Index = Index + 1;
    end
  end
  BaudManifestUpdateDisplayList(Display);
end


--First argument is the only one required
function BaudManifestGetCategory(Display, Name, ParentCategory)
  if BaudManifestCategoriesDisabled(Display)then
    UIErrorsFrame:AddMessage("Cannot create categories while the list is sorted or filtered.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    return;
  end
  if ParentCategory then
    for Key, Value in ipairs(Display.Data[ParentCategory])do
      if Value.Category and(Display.Data[Value.Category].Name==Name)then
        return Value.Category;
      end
    end
  end
  if not Name then
    Name = "New Category";
  end
  DebugMsg("Creating category: "..Name);
  tinsert(Display.Data,{Name=Name, Expanded=false});
  local Category = #Display.Data;
  tinsert(Display.Data[(ParentCategory or 1)],1,{Category = Category});
  FauxScrollFrame_SetOffset(getglobal(Display:GetName().."ScrollBar"),0);
  if not ParentCategory then
    BaudManifestUpdateDisplayList(Display);
  end
  return Category, true;
end


function BaudManifestUpdateBlocked()
  --Enabling and disabling item entries to serve this purpose would prevent items from being dropped by releasing the cursor
  local Enable = CursorItem or VirtualDrag;
  for Display = 1, 4 do
    local Frame = getglobal("BaudManifestDisplay"..Display.."ClickBlock");
    if Enable then
      Frame:Show();
    else
      Frame:Hide();
    end
  end
--  DebugMsg("Click block "..(Enable and "Enabled." or "Disabled."));
end


hooksecurefunc("PickupContainerItem",function(Bag,Slot)
  if IgnoreHooks or not CursorHasItem()then
    return;
  end
  DragItem = GetContainerItemLink(Bag, Slot);
  if(DragItem~=select(3,GetCursorInfo()))then
    DragItem = nil;
    return;
  end
  DragItem = strmatch(DragItem, ItemStringMatch);
  if not DragItem then
    return;
  end
  DragItem = {
    String = DragItem,
    Stack = select(2,GetContainerItemInfo(Bag,Slot)),
    Origin = getglobal("BaudManifestDisplay"..(((Bag == -1)or(Bag > 4))and 2 or 1)),
    Bag = Bag,
    Slot = Slot,
    IsBag = Bag and(Bag > 0)and CursorCanGoInSlot(InvBagsOffset + 1)and ContainerIDToInventoryID(Bag),
  };
  BaudManifestShowDragCount();
  DebugMsg("Item pickup from "..Bag..", "..Slot);
end);


--This hook is to correct the problem of not being able to equip a bag when it is inside the bag it is trying to replace.
--As a solution, BaudManifest will move the bag to another place in your inventory, and try to equip it again.
hooksecurefunc("PutItemInBag", function(InvID)
  if CursorHasItem()or not DragItem or not DragItem.IsBag or(InvID~=DragItem.IsBag)then
    return;
  end
  if select(3, GetContainerItemInfo(DragItem.Bag, DragItem.Slot))then
    ChatFrame1:AddMessage("Bag is locked.");
    return;
  end
  IgnoreHooks = true;
  PickupContainerItem(DragItem.Bag, DragItem.Slot);
  IgnoreHooks = nil;
  if not BaudManifestPutItem(DragItem.Origin, DragItem.Bag)then
    return;
  end
  --New location for bag should be updated by BaudManifestPutItem
  DragItem.EquipBag = true;
  DragPending = DragItem;
  DragItem = nil;
end);


hooksecurefunc("PickupBagFromSlot", function(Slot)
  if IgnoreHooks or not CursorHasItem()then
    return;
  end
  local Bag;
  if(Slot > InvBagsOffset)and(Slot <= InvBagsOffset + 4)then
    Bag = Slot - InvBagsOffset;
  elseif(Slot > BankBagsOffset)and(Slot <= BankBagsOffset + NUM_BANKBAGSLOTS)then
    Bag = Slot - BankBagsOffset + 4;
  else
    return;
  end
  DragItem = GetInventoryItemLink("player", Slot);
  if(DragItem~=select(3,GetCursorInfo()))then
    DragItem = nil;
    return;
  end
  DragItem = strmatch(DragItem, ItemStringMatch);
  if not DragItem then
    return;
  end
  DragItem = {String = DragItem, MoveBag = Bag, Origin = getglobal("BaudManifestDisplay"..(((Bag == -1)or(Bag > 4))and 2 or 1))};
  DebugMsg("Bag pickup from "..Bag);
end);


--Used for deleting multiple items at once.  Use caution!
hooksecurefunc("DeleteCursorItem",function()
  if IgnoreHooks then
    return;
  end
  BaudManifestPickupDragRemain(function()
    --This check is used as a failsafe
    local Type, _, Link = GetCursorInfo();
    if(Type~="item")then
      return;
    end
    if(DragItem.String~=strmatch(Link, ItemStringMatch))then
      DEFAULT_CHAT_FRAME:AddMessage("Baud Manifest: Item match failure.",1,0.1,0.1);
      return;
    end
    IgnoreHooks = true;
    DeleteCursorItem();
    IgnoreHooks = false;
    return true;
  end, true);
end);


--This function is used when dragging items onto the vendor window.  The values passed to it don't matter in this case.
hooksecurefunc("PickupMerchantItem",function()
  if IgnoreHooks or not VendorOpen then
    return;
  end
  BaudManifestPickupDragRemain(function()
    if VendorOpen then
      IgnoreHooks = true;
      PickupMerchantItem();
      IgnoreHooks = false;
      return true;
    end
    return false;
  end, true);
end);


function BaudManifestTradeRemain()
  if TradeOpen then
    IgnoreHooks = true;
    DropItemOnUnit("NPC");
    IgnoreHooks = false;
    return true;
  end
  return false;
end


hooksecurefunc("ClickSendMailItemButton",function()
  if IgnoreHooks or not MailOpen or CursorHasItem()then
    return;
  end
  BaudManifestPickupDragRemain(function()
    if MailOpen then
      IgnoreHooks = true;
      ClickSendMailItemButton();
      IgnoreHooks = false;
      return true;
    end
    return false;
  end);
end);


function BaudManifestDragCount_OnUpdate(self)
  local X, Y = GetCursorPosition();
  self:ClearAllPoints();
  self:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT", X - 5, Y - 27);
end


function BaudManifestVirtualDrag_OnUpdate(self)
  --Dragged categories will not be cleared for stuff that is dragged and immediately placed, but if anything is in the cursor
  --for an update, then the category is dropped.
  if GetCursorInfo()then
    BaudManifestClearVirtualDrag();
    return;
  end
  self:ClearAllPoints();
  self:SetPoint("BOTTOMRIGHT",UIParent,"BOTTOMLEFT",GetCursorPosition());
end


function BaudManifestVirtualDragClickBlock_OnClick(Button)
  if(Button=="RightButton")or not VirtualDrag.Category then
    BaudManifestClearVirtualDrag();
  else
    StaticPopup_Show("BAUDMANIFEST_DELETECATEGORY");
  end
end


function BaudManifestShowDragCount()
  local Count = DragItem.Stack;
  if DragItem.Remain then
    Count = Count + DragItem.Remain;
  end
  if not CursorHasItem()or(Count==1)and(select(8,GetItemInfo(DragItem.String))==1)then
    return;
  end
  BaudManifestDragCountText:SetText(Count);
  DebugMsg("Showing drag count of "..Count);
  BaudManifestDragCount:Show();
end


function BaudManifestBagPriority(Bag, Item)
  --nil - Item cannot go in this bag, 0 - Unknown, 1 - Any item could go here, 2 - This bag is specificly for this item type
  local BagType;
  if not Bag or(Bag == -2)then
    return 0;
  end
  if(Bag < 1)then
    return 1; --Backpack + Bank
  end
  BagType = GetInventoryItemLink("player",ContainerIDToInventoryID(Bag));
  if not BagType then
    return 0;
  end
  BagType = select(7, GetItemInfo(BagType));
  if not BagType then
    return 0;
  end
  if(BagType==BagTypes.Standard)then
    return 1;
  end
  local Holds = BagTypes[BagType];
  if Holds and Item and Holds[Item]then
    return 2;
  end
  return 0;
end


QueuedUpdateFrame:SetScript("OnUpdate", function(self)
  self:Hide();
  for Display = 1, 2 do
    if QueuedUpdate[Display]then
      if DragPending and(DragPending.Origin:GetID()==Display)and DragPending.Ready then
        if CursorHasItem()then
          DragPending = nil;
        else
          DebugMsg("Pending status finished.");
          DragItem = DragPending;
          DragPending = nil;
          DragItem.Ready = nil;
          if DragItem.MoveBag then
            --There may still be locked items in the process of being moved out of the bag.  Emptying will be repeated until
            --the bag is empty.
            BaudManifestEmptyBag(DragItem.Origin, DragItem.MoveBag, true);
          elseif DragItem.EquipBag then
            --Used for equipping bags in place of the bag that contains them
            if(strmatch(GetContainerItemLink(DragItem.Bag, DragItem.Slot)or "",ItemStringMatch)==DragItem.String)then
              IgnoreHooks = true;
              PickupContainerItem(DragItem.Bag, DragItem.Slot);
              IgnoreHooks = nil;
              PutItemInBag(DragItem.IsBag);
            end
          else
          --if the split item is picked up immediately, the client will not register that it is locked
            BaudManifestPickupDragRemain();
          end
        end
      end
      QueuedUpdate[Display] = false;
      BaudManifestUpdateList(getglobal("BaudManifestDisplay"..Display));
    end
  end
end);


ClearCursorFrame:SetScript("OnUpdate", function(self)
  self:Hide();
  if DragItem and not CursorItem then
    DebugMsg("Clearing dragged item.");
    DragItem = nil;
    BaudManifestDropLine:Hide();
  end
end);


CursorHoverFrame:SetScript("OnUpdate", function(self)
  if(DragItem and(CursorHover==DragItem.Origin)and not DragItem.MoveBag or VirtualDrag and(CursorHover==VirtualDrag.Origin))then
    local Position, Indent, _, Index = BaudManifestGetCursorPos(CursorHover);
    local ScrollBar = getglobal(CursorHover:GetName().."ScrollBar");
    BaudManifestDropLine:SetParent(CursorHover);
    BaudManifestDropLine:ClearAllPoints();
    local Y = 16 * (1 - Position);
    BaudManifestDropLine:SetPoint("TOP",ScrollBar,"TOP", 0, Y + 1);
    BaudManifestDropLine:SetPoint("LEFT",ScrollBar,"LEFT",Indent * IndentSize,0);
    BaudManifestDropLine:SetPoint("RIGHT",ScrollBar);
    if not Index then
      BaudManifestDropLine:SetPoint("BOTTOM",ScrollBar,"TOP", 0, Y - 16);
    end
    BaudManifestDropLine:SetFrameLevel(ScrollBar:GetFrameLevel() + 3);
    BaudManifestDropLine:Show();
  end
end);


function BaudManifestForEachSlot(Display, SlotFunc, BagFunc)
  local function RunBagFunc(Bag, Total)
    if(Total <= 0)or BagFunc and BagFunc(Bag)then
      return;
    end
    for Slot = 1, Total do
      if SlotFunc(Bag, Slot)then
        return true;
      end
    end
  end

  if(Display:GetID()==1)then
    for Bag = 0, 4 do
      if RunBagFunc(Bag, GetContainerNumSlots(Bag))then
        return true;
      end
    end
    --Equipped items
    if not BagFunc or not BagFunc(nil)then
      for Slot = 1, 19 do
        if SlotFunc(nil, Slot)then
          return true;
        end
      end
      for Slot = InvBagsOffset + 1, InvBagsOffset + 4 do
        if SlotFunc(nil, Slot)then
          return true;
        end
      end
    end
    --Bliz has a function to say if you have a key ring, so hey, why not?
    if HasKey()and RunBagFunc(-2, GetContainerNumSlots(-2))then
      return true;
    end
  else
    if RunBagFunc(-1, NUM_BANKGENERIC_SLOTS)then
      return true;
    end
    for Bag = 5, 4 + NUM_BANKBAGSLOTS do
      if RunBagFunc(Bag, GetContainerNumSlots(Bag))then
        return true;
      end
    end
    if not BagFunc or not BagFunc(nil)then
      for Slot = BankBagsOffset + 1, BankBagsOffset + NUM_BANKBAGSLOTS do
        if SlotFunc(nil, Slot)then
          return true;
        end
      end
    end
  end
  return false;
end


--Used to anchor Blizzard's drop down menus so they resize themselves.
local function AnchorDropdown(DropDown)
  getglobal(DropDown:GetName().."Middle"):SetPoint("RIGHT",-25,0);
  getglobal(DropDown:GetName().."Text"):SetPoint("LEFT",DropDown:GetName().."Left","RIGHT",7,0);
  DropDown.noResize = 1;
end


function BaudManifestDisplay_OnLoad(self)
  if(self:GetID() <= 2)then
    self:RegisterForDrag("LeftButton");
    tinsert(UISpecialFrames, self:GetName()); --UIChildWindows causes the window to close when opening mail, etc.
    self.ClearZero = CreateFrame("Frame", nil, self);
    self.ClearZero:SetScript("OnUpdate", BaudManifestClearZeroUpdate);
  end
  self.RefreshFrame = CreateFrame("Frame", nil, self);
  self.RefreshFrame:SetScript("OnUpdate", BaudManifestRefreshUpdate);
  self.FadeFrame = CreateFrame("Frame", nil, self);
  self.FadeFrame:SetScript("OnUpdate", BaudManifestFadeUpdate);

  self:RegisterForClicks("LeftButtonUp","RightButtonUp");

  if(self:GetID()%2==1)then
    --Remove the gap on displays with no money frame on the bottom
    getglobal(self:GetName().."FilterEditBox"):SetPoint("BOTTOM",0,26);
  end
  self.List = {};
  self.CustomFilter = nil;

  local DropDown = getglobal(self:GetName().."SortDropDown");
  AnchorDropdown(DropDown);
  BaudManifestSortDropDown_Set(DropDown, 1);

  DropDown = getglobal(self:GetName().."FilterDropDown");
  AnchorDropdown(DropDown);
  BaudManifestFilterDropDown_Set(DropDown, 1);

  BaudManifestResetButton_Update(self);

  getglobal(self:GetName().."ClickBlock"):SetFrameLevel(self:GetFrameLevel() + 4);

  --Can't be done in XML because the text must be made before the anchor frame is
  getglobal(self:GetName().."NoEntries"):SetPoint("CENTER",self:GetName().."ScrollBar");
  getglobal(self:GetName().."Text"):SetPoint("RIGHT",self:GetName().."OptionsButton","LEFT");

  local Divider = getglobal(self:GetName().."Divider");
  Divider:SetTexCoord(0,1,0.125,1,0,0,0.125,0);
  Divider:SetVertexColor(0.5,0.5,0.5);

  BaudManifestDisplay_OnSizeChanged(self);

  local Text = getglobal(self:GetName().."Text");
  if(self:GetID() > 2)then
    Text:SetPoint("TOPLEFT", 10, -10);
    return;
  end

  Text:SetPoint("TOPLEFT", 28, -10);

  local Button, Frame, Texture, BagSlot, Text, Inherit;
  local function CheckMimick(self, ...)
    HookData[self:GetName()]:SetChecked(...);
  end
  CreateFrame("CheckButton", self:GetName().."BagsButton", self, "BaudManifestBagsButtonTemplate"):SetPoint("TOPLEFT", 5, -5);

  local BagFrame = CreateFrame("Button", self:GetName().."BagsFrame", self);
  self.BagsFrame = BagFrame;
  BagFrame:Hide();
  BagFrame:SetPoint("TOPRIGHT", self, "TOPLEFT");
  local Edge = 10;
  if(self:GetID() == 1)then
    Button = CreateFrame("CheckButton", BagFrame:GetName().."BackpackButton", BagFrame, "ItemButtonTemplate");
    Button:SetPoint("TOP", 0, -Edge);
    Button:RegisterForClicks("LeftButtonUp", "RightButtonUp");
    getglobal(Button:GetName().."IconTexture"):SetTexture("Interface\\Buttons\\Button-Backpack-Up");
    Inherit = MainMenuBarBackpackButton;
    Button:SetScript("OnClick", Inherit:GetScript("OnClick"));
    Button:SetScript("OnReceiveDrag", Inherit:GetScript("OnReceiveDrag"));
    Button:SetScript("OnEnter", Inherit:GetScript("OnEnter"));
    Button:SetScript("OnLeave", Inherit:GetScript("OnLeave"));
    Texture = Button:CreateTexture();
    Texture:SetTexture("Interface\\Buttons\\CheckButtonHilight");
    Texture:SetBlendMode("ADD");
    Button:SetCheckedTexture(Texture);
    Texture:SetAllPoints();
    HookData[Inherit:GetName()] = Button;
    hooksecurefunc(Inherit, "SetChecked", CheckMimick);
    for Bag = 1, 4 do
      --Bag name length is restricted.  for this one, the ID is set automatically.
      BagSlot = CreateFrame("CheckButton", "BaudMInveBag"..(Bag - 1).."Slot", BagFrame, "BagSlotButtonTemplate");
      BagSlot:SetPoint("TOP", 0, -Edge - Bag * 39);
      getglobal(BagSlot:GetName().."ItemAnim"):UnregisterAllEvents();
      Frame = "CharacterBag"..(Bag-1).."Slot";
      HookData[Frame] = BagSlot;
      hooksecurefunc(getglobal(Frame), "SetChecked", CheckMimick);
    end
    --KEYRING BUTTON--
    Button = CreateFrame("CheckButton", BagFrame:GetName().."KeyRingButton", BagFrame);
    Button:SetWidth(39);
    Button:SetHeight(18);
    Button:SetPoint("BOTTOM", 0, Edge);
    Button:SetID(KEYRING_CONTAINER);
    Button:RegisterForClicks("LeftButtonUp", "RightButtonUp");
    Texture = Button:CreateTexture();
    Texture:SetTexture("Interface\\Buttons\\UI-Button-KeyRing");
    Texture:SetTexCoord(0.5625,0,0,0,0.5625,0.60937,0,0.60937);
    Button:SetNormalTexture(Texture);
    Texture:SetAllPoints();
    Texture = Button:CreateTexture();
    Texture:SetTexture("Interface\\Buttons\\UI-Button-KeyRing-Highlight");
    Texture:SetTexCoord(0.5625,0,0,0,0.5625,0.60937,0,0.60937);
    Button:SetHighlightTexture(Texture);
    Texture:SetAllPoints();
    Texture = Button:CreateTexture();
    Texture:SetTexture("Interface\\Buttons\\UI-Button-KeyRing-Down");
    Texture:SetTexCoord(0.5625,0,0,0,0.5625,0.60937,0,0.60937);
    Button:SetPushedTexture(Texture);
    Texture:SetAllPoints();
    Button:SetScript("OnClick", KeyRingButton:GetScript("OnClick"));
    Button:SetScript("OnReceiveDrag", KeyRingButton:GetScript("OnReceiveDrag"));
    Button:SetScript("OnEnter", KeyRingButton:GetScript("OnEnter"));
    Button:SetScript("OnLeave", KeyRingButton:GetScript("OnLeave"));
    HookData.KeyRingButton = Button;
    hooksecurefunc(KeyRingButton, "SetButtonState", function(self, ...)
      HookData[self:GetName()]:SetButtonState(...);
    end);
    BagFrame:SetWidth(Edge * 2 - 3 + 39);
    BagFrame:SetHeight(Edge * 2 - 3 + 5 * 39 + 20);
  else
    for Bag = 1, NUM_BANKBAGSLOTS do
      --Bag name length is restricted
      BagSlot = CreateFrame("Button","BaudMBankBag"..Bag, BagFrame,"BankItemButtonBagTemplate");
      BagSlot:SetID(Bag + 4);
      BagSlot:SetPoint("TOPLEFT",Edge + ((Bag - 1) % 2) * 39,-Edge - floor((Bag - 1) / 2) * 39);
      Frame = "BankFrameBag"..Bag.."HighlightFrameTexture";
      HookData[Frame] = getglobal(BagSlot:GetName().."HighlightFrameTexture");
      hooksecurefunc(getglobal(Frame), "Show", function(self)
        HookData[self:GetName()]:Show();
      end);
      hooksecurefunc(getglobal(Frame), "Hide", function(self)
        HookData[self:GetName()]:Hide();
      end);
    end
    local Button = CreateFrame("Button", "BaudManifestBankSlotPurchaseButton", BagFrame, "UIPanelButtonTemplate");
    Button:SetHeight(21);
    Button:SetPoint("BOTTOMLEFT", 7, 7);
    Button:SetPoint("RIGHT", -7, 0);
    Button:SetText(BANKSLOTPURCHASE);
    Button:SetScript("OnClick", function()
      PlaySound("igMainMenuOption");
      StaticPopup_Show("CONFIRM_BUY_BANK_SLOT");
    end);
    Button:Disable();
    Text = Button:CreateFontString(Button:GetName().."CostText", "BACKGROUND", "GameFontHighlight");
    Text:SetText("Slot:");
    Text:SetPoint("BOTTOMLEFT", Button, "TOPLEFT", 10, 2);
    Frame = CreateFrame("Frame", Button:GetName().."MoneyFrame", Button, "SmallMoneyFrameTemplate");
    Frame:SetPoint("BOTTOMLEFT", Button, "TOPLEFT", 40, 3);
    MoneyFrame_SetType("STATIC", Frame);
    Frame:Show();
    BagFrame:SetWidth(Edge * 2 - 3 + 2 * 39);
    --Height changes depending if there is a purchase button
    BagFrame.Height = Edge * 2 - 3 + ceil(NUM_BANKBAGSLOTS / 2) * 39;
  end
end


function BaudManifestOptionsDropDown_OnLoad(self)
  UIDropDownMenu_Initialize(self, BaudManifestOptionsDropDown_Initialize, "MENU");
end


local Display, Alpha;
function BaudManifestFadeUpdate(self)
  Display = self:GetParent();
  Alpha = (GetTime() - Display.FadeStart) / FadeTime;
  if Display.Closing then
    Alpha = 1 - Alpha;
    if(Alpha < 0)then
      Display.FadeStart = nil;
      Display:Hide();
      Display.Closing = nil;
      self:Hide();
      return;
    end
  elseif(Alpha > 1)then
    Display:SetAlpha(1);
    Display.FadeStart = nil;
    self:Hide();
    return;
  end
  Display:SetAlpha(Alpha);
end


function BaudManifestRefreshUpdate(self)
  if(self:GetParent().Refresh > GetTime())then
    return;
  end
  self:Hide();
  local Display = self:GetParent();
  DebugMsg("Refreshing "..Display:GetName());
  if Display.UpdateList then
    BaudManifestUpdateDisplayList(Display);
  else
    BaudManifestScrollBar_Update(Display);
  end
end


function BaudManifestClearZeroUpdate(self)
  local Display = self:GetParent();
  local Scale = Display:GetEffectiveScale();
  local X, Y = GetCursorPosition();
  X, Y = X / Scale, Y / Scale;
  if(X < Display:GetLeft())or(X > Display:GetRight())or(Y < Display:GetBottom()or(Y > Display:GetTop()))then
    BaudManifestClearZeros(Display);
    BaudManifestUpdateDisplayList(Display);
    self:Hide();
  end
end


function BaudManifestClearZeros(Display)
  if(Display:GetID() > 2)then
    return;
  end
  Display.ClearZero:Hide();
  local Index, Delete;
  for Category, Value in ipairs(Display.Data)do
    Index = 1;
    while(Value[Index])do
      if Value[Index].ItemString and not Value[Index].Count then
        if(Category==1)or(Display:GetID()==1)and(PlayerData.PutQuality[select(3,GetItemInfo(Value[Index].ItemString))or -1]==Category)then
          Display.Hash[Value[Index].ItemString] = nil;
          tremove(Value,Index);
          Index = Index - 1;
        elseif Value[Index].Show then
          Value[Index].Show = nil;
        end
      end
      Index = Index + 1;
    end
  end
end


function BaudManifestDisplay_OnShow(self)
  if self.Shown then
    return;
  end
  self.Shown = true;
  if self.FadeStart then
    return;
  end
  if(PlayerData.Fading~=false)then
    self.FadeStart = GetTime();
    self.FadeFrame:Show();
  else
    self:SetAlpha(1);
  end
  self.Closing = false;
  PlaySound("igBackPackOpen");
  local Level = self:GetFrameLevel() + 1;
  local Entry;
  for Index = 1, self.Entries do
    Entry = getglobal(self:GetName().."Entry"..Index);
    Entry:SetFrameLevel(Level);
    getglobal(Entry:GetName().."BlizItem"):SetFrameLevel(Level + 1);
    getglobal(Entry:GetName().."BlizInv"):SetFrameLevel(Level + 1);
  end
  BaudManifestClearZeros(self);
  BaudManifestUpdateDisplayList(self);
  BaudManifestDisplay_OnSizeChanged(self);
end


function BaudManifestDisplay_OnHide(self)
  if self:IsShown()then
    return;
  end
  self.Shown = nil;
  if self.Closing then
    if self.FadeStart then
      self:Show();
    end
    return;
  end
  if(PlayerData.Fading~=false)then
    self.FadeStart = GetTime();
    self.FadeFrame:Show();
    self.Closing = true;
    self:Show();
  end
  PlaySound("igBackPackClose");
  if(self:GetID()==2)and BankOpen and(PlayerData.ShowBank~=false)then
    CloseBankFrame();
  end
  self.AutoShown = nil;
end


function BaudManifestDisplay_OnSizeChanged(Display)
  local ScrollBar = getglobal(Display:GetName().."ScrollBar");
  local Entries = floor(ScrollBar:GetHeight()/16);
  if(Entries == Display.Entries)then
    return;
  end
  Display.Entries = Entries;
  local Button;
  if not Display.MaxEntries or(Display.MaxEntries < Entries)then
    for Index = (Display.MaxEntries or 0) + 1, Entries do
      Button = CreateFrame("Button",Display:GetName().."Entry"..Index,Display,"BaudManifestItemEntryTemplate");
      --Showing the texture affects the text positioning for some reason, and cannot be done in OnLoad
      getglobal(Button:GetName().."Texture"):Show();
      Button:ClearAllPoints();
      Button:SetPoint("TOP",ScrollBar,"TOP",0,16*(1-Index));
      Button:SetPoint("LEFT",ScrollBar);
      Button:SetPoint("RIGHT",ScrollBar);
      Button.Display = Display;
    end
    Display.MaxEntries = Entries;
  else
    for Index = (Entries + 1), Display.MaxEntries do
      getglobal(Display:GetName().."Entry"..Index):Hide();
    end
  end
  BaudManifestScrollBar_Update(Display)
end


function BaudManifestClearFocus()
  BaudManifestCategoryEditBox:ClearFocus();
  for Display = 1, 4 do
    getglobal("BaudManifestDisplay"..Display.."FilterEditBox"):ClearFocus();
  end
end


function BaudManifestDisplay_OnClick(Display, Button)
  DebugMsg("Click on "..Display:GetName());
  BaudManifestClearFocus();
  if(Button=="RightButton")then
    BaudManifestClearVirtualDrag();
  else
    BaudManifestDropItem(Display);
  end
end


function BaudManifestDisplay_OnDragStart(Display)
  if Display:IsMovable()then
    Display:StartMoving();
  end
end


function BaudManifestDisplay_OnDragStop(Display)
  Display:StopMovingOrSizing();
end


function BaudManifestDisplay_OnEnter(Display)
  CursorHover = Display;
  CursorHoverFrame:Show();
end


function BaudManifestDisplay_OnLeave()
  BaudManifestDropLine:Hide();
  CursorHover = nil;
  CursorHoverFrame:Hide();
end


function BaudManifestResize(self)
  local Display = self:GetParent();
  if Display.Data.LockSize or not Display:IsResizable()then
    BaudManifestDisplay_OnDragStart(Display);
    return;
  end
  BaudManifestClearFocus();
  local AnchorPoint = strupper(strsub(self:GetName(),strlen(Display:GetName().."Resize")+1));
  Display:StartSizing(AnchorPoint);
end


function BaudManifestStopResize(self)
  self:GetParent():StopMovingOrSizing();
end


function BaudManifestItemEntry_OnLoad(self)
  self:RegisterForClicks("LeftButtonUp","RightButtonUp");
  self:RegisterForDrag("LeftButton");
  self.UpdateTooltip = BaudManifestItemEntry_OnEnter;

  --this prevents errors from occuring with Bliz's split stack frame
  getglobal(self:GetName().."BlizItem").count = 0;
end


--Used for replacing Blizzard's call for opening the split stack frame
hooksecurefunc("OpenStackSplitFrame", function(maxStack, parent, anchor, anchorTo)
  if not StackSplitFrame:IsShown()and (strsub(parent:GetName(), 1, 12)=="BaudManifest")then
    BaudManifestItemEntry_OnDragStart(parent:GetParent());
  end
end);


function BaudManifestItemEntry_OnClick(Frame, MouseButton)
  local Display = Frame:GetParent();
  local ItemInfo = Display.Data[Frame.Category][Frame.ItemIndex];
  BaudManifestClearFocus();

  if ItemInfo.Category then
    if(MouseButton=="RightButton")then
      DebugMsg("Edit category name");
      BaudManifestCategoryEditBox:Hide();
      local Text = getglobal(Frame:GetName().."Text");
      Text:Hide();
      --Unless the parent is set, the edit box can fall behind the frame
      BaudManifestCategoryEditBox:SetParent(Frame);
      BaudManifestCategoryEditBox:SetFrameLevel(Frame:GetFrameLevel() + 1);
      BaudManifestCategoryEditBox:SetPoint("LEFT",Text);
      BaudManifestCategoryEditBox:SetPoint("RIGHT",Text);
      BaudManifestCategoryEditBox.Display = Display;
      BaudManifestCategoryEditBox.Category = Display.Data[ItemInfo.Category];
      BaudManifestCategoryEditBox:SetText(Display.Data[ItemInfo.Category].Name);
      BaudManifestCategoryEditBox:Show();
      BaudManifestCategoryEditBox:SetFocus();
      BaudManifestCategoryEditBox:HighlightText();
    else
      DebugMsg("Edit category toggle expand");
      Display.Data[ItemInfo.Category].Expanded = not Display.Data[ItemInfo.Category].Expanded;
      BaudManifestUpdateDisplayList(Display);
    end
    return;
  end

  if(MouseButton=="LeftButton")then
    if HandleModifiedItemClick(select(2,GetItemInfo(ItemInfo.ItemString..":"..(ItemInfo.SuffixFactor or 0))))then
      return;
    end
    if IsModifiedClick("SPLITSTACK")then
      BaudManifestItemEntry_OnDragStart(Frame);
    end
    --Support for another addon (I don't think it's needed anymore)
    --[[if AltClickToAddItem and(strsub(this:GetName(),-8)=="BlizItem")and IsAltKeyDown()then
      ContainerFrameItemButton_OnModifiedClick(Button);
    end]]
  end
end


local ItemTooltipAppend;
function BaudManifestItemEntry_OnEnter(Button, NotUpdate)
  if NotUpdate then
    BaudManifestItemHighlight:SetParent(Button);
    BaudManifestItemHighlight:SetFrameLevel(Button:GetFrameLevel() + 1);
    BaudManifestItemHighlight:SetAllPoints();
    BaudManifestItemHighlight:Show();
  end
  local Display = Button:GetParent();
  local ItemInfo = Display.Data[Button.Category][Button.ItemIndex];
  if not ItemInfo.ItemString then
    Button.updateTooltip = nil;
    return;
  end
  if NotUpdate then
    --if this is updated repeatedly, it causes flickering
    BaudManifestTooltipIconTexture:SetTexture(select(10,GetItemInfo(ItemInfo.ItemString)));
    CooldownFrame_SetTimer(BaudManifestTooltipIconCooldown, GetItemCooldown(ItemInfo.ItemString));
    BaudManifestTooltipIcon:Show();

    BaudManifestItemHighlight:SetParent(Button);
    BaudManifestItemHighlight:SetFrameLevel(Button:GetFrameLevel() + 1);
  end
  if(Button:GetCenter() < UIParent:GetCenter())then
    GameTooltip:SetOwner(Button, "ANCHOR_RIGHT");
  else
    GameTooltip:SetOwner(Button, "ANCHOR_LEFT");
  end
  local SkipReset;
  local HasCooldown, RepairCost;
  if(Display:GetID() > 2)or(Display:GetID()==2)and not BankOpen or not ItemInfo.Slot then
    GameTooltip:SetHyperlink(ItemInfo.ItemString..":"..(ItemInfo.SuffixFactor or 0));
  elseif not ItemInfo.Bag then
    _, HasCooldown, RepairCost = GameTooltip:SetInventoryItem("player",ItemInfo.Slot);
  elseif(ItemInfo.Bag==-1)then
    _, HasCooldown, RepairCost = GameTooltip:SetInventoryItem("player",BankButtonIDToInvSlotID(ItemInfo.Slot));
  elseif(ItemInfo.Bag==-2)then
    _, HasCooldown, RepairCost = GameTooltip:SetInventoryItem("player",KeyRingButtonIDToInvSlotID(ItemInfo.Slot));
  else
    HasCooldown, RepairCost = GameTooltip:SetBagItem(ItemInfo.Bag,ItemInfo.Slot);
  end

  if InRepairMode()and RepairCost and(RepairCost > 0)and IsDisplayAccessable(Display)then
    GameTooltip:AddLine(REPAIR_COST, "", 1, 1, 1);
    SetTooltipMoney(GameTooltip, RepairCost);
    GameTooltip:Show();
  elseif IsControlKeyDown()then
    ShowInspectCursor();
  elseif ItemInfo.Bag and ItemInfo.Slot and MerchantFrame:IsShown()and(MerchantFrame.selectedTab==1)and IsDisplayAccessable(Display)then
    ShowContainerSellCursor(ItemInfo.Bag,ItemInfo.Slot);
  else
    ResetCursor();
  end

  if NotUpdate then
    ItemTooltipAppend = "";
    if ItemInfo.Equipped then
      ItemTooltipAppend = "Equipped\n";
    end

    if BaudManifestCategoriesDisabled(Display)and Display.Data[Button.Category].Path then
      ItemTooltipAppend = ItemTooltipAppend.."Category: "..Display.Data[Button.Category].Path.."\n";
    end

    if(Display.Name=="All Characters")then
      local CharList = {};
      local Count;
      local Bank;
      for Char, Value in pairs(RealmData)do
        if(Char~=Display.Name)then
          Count, Bank = 0, 0;
          for Display, Value in ipairs(Value)do
            for Category, Value in ipairs(Value)do
              for Index, Value in ipairs(Value)do
                if(Value.ItemString==ItemInfo.ItemString)and Value.Count then
                  Count = Count + Value.Count;
                  if(Display==2)then
                    Bank = Bank + Value.Count;
                  end
                end
              end
            end
          end
          if(Count > 0)then
            tinsert(CharList,{Name=Char, Count=Count, Bank = Bank});
          end
        end
      end
      if(#CharList > 0)then
        table.sort(CharList,function(a,b) return(a.Count > b.Count);end);
        for Key, Value in ipairs(CharList)do
          ItemTooltipAppend = ItemTooltipAppend..Value.Name..": "..Value.Count;
          if(Value.Bank > 0)then
            if(Value.Bank==Value.Count)then
              ItemTooltipAppend = ItemTooltipAppend.." (Bank)";
            else
              ItemTooltipAppend = ItemTooltipAppend.." ("..Value.Bank.." Bank)";
            end
          end
          ItemTooltipAppend = ItemTooltipAppend.."\n";
        end
      end
    end
    if(ItemTooltipAppend=="")then
      ItemTooltipAppend = nil;
    else
      ItemTooltipAppend = strsub(ItemTooltipAppend, 1, -2);
    end
  end
  if ItemTooltipAppend then
    GameTooltip:AddLine(ItemTooltipAppend);
    GameTooltip:Show();
  end
end


function BaudManifestItemEntry_OnLeave()
  GameTooltip:Hide();
  BaudManifestItemHighlight:Hide();
  BaudManifestTooltipIcon:Hide();
  ResetCursor();
end


function  BaudManifestDisplayMoney_OnEnter(self)
  local Display = self:GetParent();
  if(Display.Name~="All Characters")then
    return;
  end
  local CharList = {};
  for Char, Value in pairs(RealmData)do
    if(Char~=Display.Name)then
      tinsert(CharList,{Name=Char, Gold=floor(Value.Money / 10000)})
    end
  end
  GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT");
  GameTooltip:AddLine("Breakdown:",1,1,1);
  table.sort(CharList,function(a,b) return(a.Gold > b.Gold);end);
  for Key, Value in ipairs(CharList)do
    GameTooltip:AddLine(Value.Name..": "..Value.Gold.." gold");
  end
  GameTooltip:Show();
end


function  BaudManifestDisplayMoney_OnLeave()
  GameTooltip:Hide();
end


function BaudManifestSplitStackHandler(Button, Split)
  DragItem = {String = SplitItem, Origin = SplitOrigin, Remain = Split};
  BaudManifestPickupDragRemain();
end


--Quick split stops the addon from placing the item into an empty slot before moving it.  Best for selling and deleting.
function BaudManifestPickupDragRemain(MoveMethod, QuickSplit)
  if not DragItem or not DragItem.Remain then
    return;
  end
  if MoveMethod then
    if DragItem.MoveMethod and MoveMethod then
      DragItem = nil;
      ClearCursor();
      DebugMsg("PickupDragRemain:  Cannot override existing move method.");
      return;
    end
    DragItem.MoveMethod = MoveMethod;
    DragItem.QuickSplit = QuickSplit;
  end
  local BestBag, BestSlot, BestCount, BestPriority, EmptyBag, EmptySlot, EmptyPriority;
  local Link, Count, Locked, Priority;
  local ItemString = DragItem.String..":";
  local ItemID = strmatch(DragItem.String, "^item:(%d+):");
  if ItemID then
    ItemID = tonumber(ItemID);
  end
  while true do
    if CursorHasItem()then
      DragItem = nil;
      ClearCursor();
      DebugMsg("PickupDragRemain:  Cursor has item already - aborting.");
      return;
    end
    if(DragItem.MoveMethod==true)and not BankOpen then
      DragItem = nil;
      ClearCursor();
      DebugMsg("PickupDragRemain:  Bank is not open - aborting.");
      return;
    end
    BestSlot, EmptyPriority = nil, 0;

    BaudManifestForEachSlot(DragItem.Origin,function(Bag, Slot)
      if Bag then
        Link = GetContainerItemLink(Bag,Slot);
        _, Count, Locked = GetContainerItemInfo(Bag, Slot);
      else
        Link = GetInventoryItemLink("player",Slot);
        Count = 1;
        Locked = IsInventoryItemLocked(Slot);
      end
      if Link then
        --This chooses stacks of the most appropriate size
        if not Locked and strfind(Link,ItemString,0,true)and(not BestSlot
        or(BestCount < DragItem.Remain)and(Count > BestCount)
        or(BestCount > DragItem.Remain)and(Count >= DragItem.Remain)and(Count < BestCount))then
          BestBag, BestSlot, BestCount = Bag, Slot, Count;
        end
      elseif(Priority > EmptyPriority)then
        --Empty slots are used for splitting items before moving them
        EmptyBag, EmptySlot, EmptyPriority = Bag, Slot, Priority;
      end
    end,
    function(Bag)
      Priority = BaudManifestBagPriority(Bag, ItemID);
      --Even priority 0 bags are still scanned because the keyring will always be priority 0
    end);

    if not BestSlot then
      DragItem = nil;
      DebugMsg("Could not find item for drag - aborting.");
      return;
    end
    --if the stack size isn't larger than what's requested, the whole thing is moved
    if(BestCount <= DragItem.Remain)then
      if not BestBag then
        IgnoreHooks = true;
        PickupInventoryItem(BestSlot);
        IgnoreHooks = false;
        if(DragItem.MoveMethod==true)then
          BaudManifestPutItem(getglobal("BaudManifestDisplay"..(DragItem.Origin:GetID() % 2 + 1)));
        end
      elseif(DragItem.MoveMethod==true)then
        UseContainerItem(BestBag, BestSlot);  --Transfer to bank
      else
        IgnoreHooks = true;
        PickupContainerItem(BestBag, BestSlot);
        IgnoreHooks = false;
      end
      if(not BestBag or(DragItem.MoveMethod~=true))and not CursorHasItem()then
        DebugMsg("Item move failed - aborting.");
        DragItem = nil;
        return;
      end
      DragItem.Stack = BestCount;
      DragItem.Remain = DragItem.Remain - BestCount;
      DebugMsg("Picking up stack of "..BestCount.." from "..(BestBag or "n/a")..","..BestSlot.." ("..DragItem.Remain.." remain)");
      if(DragItem.Remain <= 0)then
        DragItem.Remain = nil;
      end
    else
      DebugMsg("Splitting "..DragItem.Remain.." from stack at "..BestBag..","..BestSlot);
      IgnoreHooks = true;
      SplitContainerItem(BestBag, BestSlot, DragItem.Remain);
      IgnoreHooks = false;
      if not CursorHasItem()then
        DebugMsg("Failure to split stack.");
        DragItem = nil;
        return;
      end
      if not DragItem.QuickSplit and(EmptyPriority > 0)then
        --When a stack must be split, the moving procedure will continue after the move is finished
        DebugMsg("Placing split item at "..EmptyBag..","..EmptySlot);
        IgnoreHooks = true;
        PickupContainerItem(EmptyBag, EmptySlot);
        IgnoreHooks = false;
        DragPending = DragItem;
        DragItem = nil;
        return;
      else
        if DragItem.QuickSplit then
          DebugMsg("Don't need to place split item.");
        elseif(EmptyPriority <= 0)then
          UIErrorsFrame:AddMessage("No room to split item.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
        end
        --DragItem.Split = true;
        DragItem.Stack = DragItem.Remain;
        DragItem.Remain = nil;
        if(DragItem.MoveMethod==true)then
          BaudManifestPutItem(getglobal("BaudManifestDisplay"..(DragItem.Origin:GetID() % 2 + 1)));
        end
      end
    end
    --if MoveMethod is set to move between displays, no item will be in the cursor unless the item cannot be moved by using it
    if(not BestBag or(DragItem.MoveMethod~=true))and not CursorHasItem()then
      DebugMsg("Item move failed - aborting.");
      DragItem = nil;
      return;
    end
    if not DragItem.MoveMethod then
      BaudManifestShowDragCount();
      return;
    else
      if(DragItem.MoveMethod==true)and CursorHasItem()or(DragItem.MoveMethod~=true)and not DragItem.MoveMethod()or
      not DragItem or not DragItem.Remain then
        DragItem = nil;
        ClearCursor();
        return;
      end
    end
  end
end


function BaudManifestItemEntry_OnDragStart(Button)
  BaudManifestClearFocus();
  local Display = Button:GetParent();
  if not BaudManifestDropItem(Display)then
    return;
  end
  local ItemInfo = Display.Data[Button.Category][Button.ItemIndex];
  if ItemInfo.Category then
    DebugMsg("Picking up category "..ItemInfo.Category);
    VirtualDrag = {Category = ItemInfo.Category, Origin = Display};
    BaudManifestVirtualDragIcon:SetTexture("Interface\\Icons\\INV_Scroll_03");
    BaudManifestVirtualDragIcon:SetAlpha(0.6);
    ClearCursor();  --Clear any spells and such in the cursor
    PlaySound("INTERFACESOUND_CURSORGRABOBJECT");
    BaudManifestVirtualDrag:Show();
    BaudManifestUpdateBlocked();
    return;
  end
  if ItemInfo.Count and(Display:GetID() <= 2)and((Display~=BaudManifestDisplay2)or BankOpen)then
    if(ItemInfo.Count > 1)and(IsShiftKeyDown()or IsAltKeyDown())then
      SplitItem = ItemInfo.ItemString;
      SplitOrigin = Display;
      if IsShiftKeyDown()and IsControlKeyDown()then
        BaudManifestSplitStackHandler(nil,ItemInfo.Count);
      elseif IsAltKeyDown()then
        BaudManifestSplitStackHandler(nil,1);
      else
        Button.SplitStack = BaudManifestSplitStackHandler;
        OpenStackSplitFrame(ItemInfo.Count, Button, "BOTTOMRIGHT", "TOPRIGHT");
      end
      return;
    end
    local ItemString = ItemInfo.ItemString;
    DebugMsg("Searching for "..ItemString);
    local Smallest = IsControlKeyDown();
    local BestBag, BestSlot, BestStack, Count, Locked, Link;
    BaudManifestForEachSlot(Display,
      function(Bag, Slot)
        if Bag then
          Link = GetContainerItemLink(Bag,Slot);
          _, Count, Locked = GetContainerItemInfo(Bag,Slot);
        else
          Link = GetInventoryItemLink("player",Slot);
          Count = 1;
          Locked = IsInventoryItemLocked(Slot);
        end
        if Link and strfind(Link,ItemString,0,true)and not Locked and(Bag or not BestBag)and(not BestBag or Smallest and
        (Count < BestStack)or not Smallest and(Count > BestStack))then
          BestBag, BestSlot, BestStack = Bag, Slot, Count;
        end
      end
    );
    if not BestStack then
      DebugMsg("Failed to find dragged item.");
    else
      if BestBag then
        --Drag item will automaticaly be set up since ignore hooks is not set
        PickupContainerItem(BestBag, BestSlot);
      else
        --Equipped items that are dragged from within the list are marked with the origin as the list so that when they are dropped,
        --they will be moved instead of placed in your inventory.  if they are dragged from the paper doll and dropped, they will
        --be unequipped.
        PickupInventoryItem(BestSlot);
        DragItem = {String = ItemString, Stack = 1, Origin = Display};
        --BaudManifestShowDragCount();
      end
      if CursorHasItem()then
        DebugMsg("Picked up item.");
        return;
      else
        DebugMsg("Failed to pickup item.");
      end
    end
  end
  DebugMsg("Virtual item pickup");
  VirtualDrag = {ItemString = ItemInfo.ItemString, Origin = Display};
  BaudManifestVirtualDragIcon:SetTexture(select(10,GetItemInfo(ItemInfo.ItemString))or "Interface\\Icons\\INV_Misc_QuestionMark");
  BaudManifestVirtualDragIcon:SetAlpha(0.6);
  ClearCursor();  --Clear any spells and such in the cursor
  PlaySound("INTERFACESOUND_CURSORGRABOBJECT");
  BaudManifestVirtualDrag:Show();
  BaudManifestUpdateBlocked();
end


hooksecurefunc("SplitContainerItem",function(Bag, Slot, Amount)
  if IgnoreHooks then
    return;
  end
  SplitHoldItem = strmatch(GetContainerItemLink(Bag, Slot)or "",ItemStringMatch);
  SplitHoldExpire = GetTime() + 120;
end);


--This function determines where dropped items will go, aswell as where the drop line appears while dragging
function BaudManifestGetCursorPos(Display)
  local ScrollBar = getglobal(Display:GetName().."ScrollBar");
  local Position = (select(2,GetCursorPosition()) / (Display:GetEffectiveScale()) - ScrollBar:GetTop()) / -16 + 1;
  if(Position < Display.MinEntry)then
    Position = Display.MinEntry;
  end
  if(Position > Display.MaxEntry + 1)then
    Position = Display.MaxEntry + 1;
  end
  local DisplayIndex = floor(Position) + FauxScrollFrame_GetOffset(ScrollBar);
  if(DisplayIndex >= #Display.List + 1)then
    if(Position > #Display.List + 1)then
      Position = #Display.List + 1;
    end
    return Position, 0, 1, (#Display.Data[1] + 1);
  end
  local DisplayInfo = Display.List[DisplayIndex];
  local ItemIndex = DisplayInfo.Index;
  local ItemInfo = Display.Data[DisplayInfo.Category][ItemIndex];
  local Remain = Position % 1;
--[[  if ItemInfo.Category and(Remain > 0.25)and
  ((Remain < 0.75)or Display.List[DisplayIndex+1]and Display.List[DisplayIndex+1].Category==ItemInfo.Category)then
    return floor(Position + 1), DisplayInfo.Indent + 1, ItemInfo.Category, 1;]]
  if ItemInfo.Category and(Remain > 0.25)and
  ((Remain < 0.75)or Display.List[DisplayIndex+1]and Display.List[DisplayIndex+1].Category==ItemInfo.Category)then
    return floor(Position), DisplayInfo.Indent, ItemInfo.Category;

  elseif(Remain >= 0.5)then
    ItemIndex = ItemIndex + 1;
  end
  Position = floor(Position + 0.5);

  return Position, DisplayInfo.Indent, DisplayInfo.Category, ItemIndex;
end


function BaudManifestMoveListItem(Display)
  if(Display.Sort~=1)then
    UIErrorsFrame:AddMessage("List items cannot be arranged while the list is sorted.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    return;
  end

  local _, _, NewCategory, NewIndex = BaudManifestGetCursorPos(Display);
  DebugMsg("Moving item to category "..NewCategory..", index "..(NewIndex or "-1"));

  --Recursive function for seeing if a category drop location is within itself
  local function FindCategory(Source)
    if(Source == NewCategory)then
      return true;
    end
    DebugMsg("Searching category #"..Source);
    for Key, Value in ipairs(Display.Data[Source])do
      if Value.Category and FindCategory(Value.Category)then
        return true;
      end
    end
  end

  if VirtualDrag and VirtualDrag.Category and FindCategory(VirtualDrag.Category)then
    UIErrorsFrame:AddMessage("You cannot place a category within itself.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    return;
  end

  local OldItem;

  --Executed as a function so I can break out of both loops at once
  local function GetOldItem()
    for Category, Value in ipairs(Display.Data)do
      for ItemIndex, Value in ipairs(Value)do
        if Value.ItemString and(DragItem and(DragItem.String==Value.ItemString)or VirtualDrag and(VirtualDrag.ItemString==Value.ItemString))
        or Value.Category and VirtualDrag and(VirtualDrag.Category==Value.Category)then
          OldItem = Value;
          tremove(Display.Data[Category],ItemIndex);
          if NewIndex and(NewCategory==Category)and(NewIndex > ItemIndex)then
            NewIndex = NewIndex - 1;
          end
          return;
        end
      end
    end
  end
  GetOldItem();

  if not OldItem then
    DebugMsg("Failed to find move item");
    return;
  end
  if NewIndex then
    tinsert(Display.Data[NewCategory],NewIndex,OldItem);
  else
    tinsert(Display.Data[NewCategory],OldItem);
  end
  BaudManifestUpdateDisplayList(Display);
end


--Places the item in a free slot
function BaudManifestPutItem(Display, SkipBag)
  local Type, ItemID = GetCursorInfo();
  if(Type~="item")then
    DebugMsg("Cursor is not carrying an item.");
    return;
  end
  local BestBag, BestSlot, BestPriority, Priority;
  BaudManifestForEachSlot(Display,
    function(Bag, Slot)
      if Bag then
        if Priority and(not BestPriority or(Priority > BestPriority))and not GetContainerItemLink(Bag, Slot)then
          BestBag, BestSlot, BestPriority = Bag, Slot, Priority;
        end
      end
    end,
    function(Bag)
      if DragItem and(Bag==DragItem.MoveBag)or(Bag==SkipBag)then
        return true;
      end
      Priority = BaudManifestBagPriority(Bag, ItemID);
      return(Priority < 1);
    end
  );
  if not BestBag then
    DebugMsg("Failed to find empty space.  Dropping item into bag to likely give an error.");
    if(Display:GetID()==1)then
      PutItemInBackpack();
    else
      BestBag = ContainerIDToInventoryID(5);
      if not CursorCanGoInSlot(BestBag)then
        PutItemInBag(BestBag);
      end
    end
    return;
  end

  DebugMsg("Dropping item at "..BestBag..", "..BestSlot..", priority "..BestPriority);
  --PutKeyInKeyring() is not an API function and just wastes more processing power
  --PutItemInBag() can be a problem if the item being dropped happens to be a bag.
  if(BestBag==0)then
    PutItemInBackpack();
  else
    local InvID = (BestBag > 0)and ContainerIDToInventoryID(BestBag);
    if InvID and CursorCanGoInSlot(InvID)then
      PickupContainerItem(BestBag, BestSlot);
    else
      PutItemInBag(InvID);
    end
  end
  --This is needed by the PutItemInBag hook
  if DragItem then
    DragItem.Bag, DragItem.Slot = BestBag, BestSlot;
  end
  return true;
end


function BaudManifestDropItem(Display)
  if not VirtualDrag and not CursorHasItem()then
    return true;
  end

  if VirtualDrag and(VirtualDrag.Origin~=Display)then
    UIErrorsFrame:AddMessage((VirtualDrag.Category and "Categories" or "That item").." cannot be moved between displays.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    BaudManifestClearVirtualDrag();
    return;
  end
  if VirtualDrag or DragItem and not DragItem.MoveBag and(DragItem.Origin==Display)then
    --Dragging an item around in its own container will only move the item in the list
    BaudManifestMoveListItem(Display);
    BaudManifestClearVirtualDrag();
    ClearCursor();
    return;
  end
  if(Display:GetID() > 2)then
    UIErrorsFrame:AddMessage("Cannot put items into other char's manifests.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    return;
  elseif(Display:GetID()==2)and not BankOpen then
    UIErrorsFrame:AddMessage("Cannot put items in the bank while the bank is closed.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    return;
  end
  if DragItem and DragItem.Origin then
    if DragItem.MoveBag then
      --Problem if player drops the unequipped bag in a different display
      BaudManifestEmptyBag(DragItem.Origin, DragItem.MoveBag, true);
      return;
    end
    ClearCursor();
    --Moving items between the bank and inventory is done by using them
    DragItem.Remain = (DragItem.Remain or 0) + DragItem.Stack;
    BaudManifestPickupDragRemain(true);
    return;
  end
  BaudManifestPutItem(Display);
end


local SpecialBags = {};
local NumSpecials;

function BaudManifestReadSpecials(Display, Skip)
  NumSpecials = 0;
  local Empties, SpecialBag, BagType, Link;
  BaudManifestForEachSlot(Display, function(Bag, Slot)
    Link = GetContainerItemLink(Bag, Slot);
    if Link then
      return;
    end
    if not SpecialBag then
      NumSpecials = NumSpecials + 1;
      if SpecialBags[NumSpecials]then
        SpecialBag = SpecialBags[NumSpecials];
      else
        SpecialBag = {};
        SpecialBag.Empty = {};
        SpecialBags[NumSpecials] = SpecialBag;
      end
      Empties = SpecialBag.Empty;
      SpecialBag.Type = BagType;
      SpecialBag.Index = Bag;
      SpecialBag.Free = 1;
      if not BagTypes[BagType]then
        BagTypes[BagType] = {};
      end
    else
      SpecialBag.Free = SpecialBag.Free + 1;
    end
    Empties[SpecialBag.Free] = Slot;
  end,
  function(Bag)
    if not Bag or(Bag < 1)or(Bag==Skip)or DragItem and(DragItem.MoveBag==Bag)then
      return true;
    end
    BagType = GetInventoryItemLink("player",ContainerIDToInventoryID(Bag));
    BagType = BagType and select(7, GetItemInfo(BagType));
    if not BagType or(BagType==BagTypes.Standard)then
      return true;
    end
    SpecialBag = nil;
  end);
end

local AddItems = {};

local function CheckClear()
end


local Empty, Slots;
local PickupBag, PickupSlot, PlaceBag, PlaceSlot;
local ItemInfo, Link, ItemString, ItemID, Suffix, UniqueID, Count, Quality, MaxCount, Locked;
local Found, InvID, CountSlots;
local CursorBusy;
local BagType, SpecialBag;
local Hash;
local function UpdateSlotFunc(Bag, Slot)
  if CountSlots then
    Slots = Slots + 1;
  end
  if Bag then
    Link = GetContainerItemLink(Bag, Slot);
    _, Count, Locked = GetContainerItemInfo(Bag, Slot);
  else
    Link = GetInventoryItemLink("player", Slot);
    Count = 1; --GetInventoryItemCount("player", Slot);
    Locked = IsInventoryItemLocked(Slot);
  end
  if not Link then
    if CountSlots then
      Empty = Empty + 1;
    end
    return;
  end
  ItemString, ItemID, Suffix, UniqueID = strmatch(Link, ItemStringMatch);
  if not ItemString then
    return;
  end
  ItemID = tonumber(ItemID);
  if SpecialBag then
    if not SpecialBag[ItemID]then
      SpecialBag[ItemID] = true;
    end
  elseif CountSlots and not CursorBusy and PlayerData.FillSpecials then
    for Index = 1, NumSpecials do
      if BagTypes[SpecialBags[Index].Type][ItemID]and(SpecialBags[Index].Free > 0)then
        SpecialBags[Index].Free = SpecialBags[Index].Free - 1;
        IgnoreHooks = true;
        PickupContainerItem(Bag, Slot);
        InvID = ContainerIDToInventoryID(SpecialBags[Index].Index);
        if CursorCanGoInSlot(InvID)then
          PickupContainerItem(SpecialBags[Index].Index, SpecialBags[Index].Empty[1]);
        else
          PutItemInBag(InvID);
        end
        IgnoreHooks = false;
        tremove(SpecialBags[Index].Empty, 1);
        Locked = true;
        break;
      end
    end
  end
  Suffix = tonumber(Suffix);
  MaxCount = select(8,GetItemInfo(ItemString));
  ItemInfo = Hash[ItemString];
  if ItemInfo then
    if(not ItemInfo.Slot or not ItemInfo.Equipped and not Bag)then
      ItemInfo.Equipped = not Bag and true or nil;
    end
    if ItemInfo.Slot then
      ItemInfo.Count = ItemInfo.Count + Count;
      ItemInfo.Slots = (ItemInfo.Slots or 1) + 1;
    else
      ItemInfo.Count = Count;
      ItemInfo.Slots = nil;
      ItemInfo.ICount = nil;
    end

    if not ItemInfo.Slot or Bag and not ItemInfo.Bag or not Locked and ItemInfo.Locked or not Locked and(Count > ItemInfo.Stack)then
      ItemInfo.Bag = Bag;
      ItemInfo.Slot = Slot;
      ItemInfo.Stack = Count;
      ItemInfo.Locked = Locked;
    end
  else
    ItemInfo = {ItemString=ItemString,Count=Count,Bag=Bag,Slot=Slot,Locked=Locked,Stack=Count};
    Hash[ItemString] = ItemInfo;
    if not Bag then
      ItemInfo.Equipped = true;
    end
    tinsert(AddItems, ItemInfo);
  end

  --When the item suffix is negative, that means there is a scaling factor stored in the first 16 bits of the UniqueID.
  --Without this scaling factor, the item's stats will not display correctly if a tooltip is created from the info.
  --As of patch 2.1, unless the remaining bits of the UniqueID are kept, linked items will disconnect the player.
  if ItemInfo.SuffixFactor or(Suffix < 0)then
    ItemInfo.SuffixFactor = (Suffix < 0)and UniqueID or nil; --bit.band(tonumber(UniqueID),65535);
  end

  --MaxCount can register nil here
  if not DisableCombine and Bag and not CursorBusy and(PlayerData.Combine~=false)and MaxCount and(Count < MaxCount)and not Locked and(SplitHoldItem~=ItemString)and(not DragPending or(DragPending.String~=ItemString))then
    if not ItemInfo.ICount then
      ItemInfo.ICount = Count;
      ItemInfo.IBag = Bag;
      ItemInfo.ISlot = Slot;
    else
      PickupBag, PickupSlot = Bag, Slot;
      PlaceBag, PlaceSlot = ItemInfo.IBag, ItemInfo.ISlot;
      --Combine two incomplete stacks by moving the larger one onto the smaller
      if(Count < ItemInfo.ICount)then
        PickupBag, PickupSlot, PlaceBag, PlaceSlot = PlaceBag, PlaceSlot, PickupBag, PickupSlot
      end
      ItemInfo.ICount = nil;
      DebugMsg("Item is "..(Locked and "locked" or "unlocked"));
      ClearCursor();
      DebugMsg("Combining incomplete stacks.");
      IgnoreHooks = true;
      PickupContainerItem(PickupBag, PickupSlot);
      if CursorHasItem()then
        PickupContainerItem(PlaceBag, PlaceSlot);
        ClearCursor();
      end
      IgnoreHooks = false;
    end
  end
end

local function UpdateBagFunc(Bag)
  SpecialBag = nil;
  CountSlots = nil;
  if Bag and(Bag~=-2)then
    if(Bag < 1)then
      CountSlots = true;
    else
      Link = GetInventoryItemLink("player",ContainerIDToInventoryID(Bag));
      SpecialBag = Link and select(7, GetItemInfo(Link));
      if SpecialBag then
        if(SpecialBag==BagTypes.Standard)then
          SpecialBag = nil;
          CountSlots = true;
        else
          SpecialBag = BagTypes[SpecialBag];
        end
      end
    end
  end
end

function BaudManifestUpdateList(Display)
  if(Display:GetID() > 2)then
    --This should never happen, so it's a red flag if it does
    error("Attempt to update manifest of another character.");
    return;
  end
  if QueuedUpdate[Display:GetID()]then
    return;
  end
  if(Display:GetID()==2)and not BankOpen then
    DebugMsg("Not updating the bank box because it is not open.");
    return;
  end
  if SplitHoldItem and(GetTime() > SplitHoldExpire)then
    SplitHoldItem = nil;
  end
  if DisableCombine and(GetTime() > DisableCombine)then
    DisableCombine = nil;
  end
  DebugMsg("Updating item list "..Display:GetID()..".");

  Hash = Display.Hash;
  --if the count is cleared here, an error during the update process could cause the count to not increase, and the items could be erased.
  for Category, Value in pairs(Hash)do
    Value.Slot = nil;
  end

  for Key, Value in pairs(AddItems)do
    AddItems[Key] = nil;
  end
  BaudManifestReadSpecials(Display);

  Empty, Slots = 0, 0;
  CursorBusy = GetCursorInfo();
  BaudManifestForEachSlot(Display, UpdateSlotFunc, UpdateBagFunc);
  --Set counts of items that were not found to 0.  Safest to be done after, in case of errors occuring, causing the scan process to end prematurely.
  for Key, Item in pairs(Hash)do
    if Item.Count and not Item.Slot then
      ItemID = strmatch(Item.ItemString, "item:(%d+):")or true; --Just to make sure this is never nil
      for Key, NewItem in ipairs(AddItems)do
        if(ItemID==strmatch(NewItem.ItemString, "item:(%d+):"))then
          --To avoid having to find the item in the list, just copy the info over.  The NewItem table is discarded after.
          --if there's a marker for this item in the other display (probably bank), then it should be updated too
          local Hash2 = PlayerHash[Display:GetID() % 2 + 1];
          local Item2 = Hash2[Item.ItemString];
          if Item2 and not Item2.Count then
            Item2.ItemString = NewItem.ItemString;
            Hash2[Item.ItemString] = nil;
            Hash2[NewItem.ItemString] = Item2;
          end
          Hash[Item.ItemString] = nil;
          Hash[NewItem.ItemString] = Item;  --This value was NewItem, but needs to be changed to point to the copy
          for Key, Value in pairs(Item)do
            Item[Key] = nil;
          end
          for Key, Value in pairs(NewItem)do
            Item[Key] = Value;
          end
          tremove(AddItems, Key);
          ItemID = nil;
          break;
        end
      end
      if ItemID then
        Item.Count = nil;
        Item.Equipped = nil;
        Item.Slots = nil;
        Item.Show = true;
        Display.ClearZero:Show();
      end
    end
  end
  for Key, Value in ipairs(AddItems)do
    Quality = select(3, GetItemInfo(Value.ItemString));
    tinsert(Display.Data[(Display:GetID()==1)and Quality and PlayerData.PutQuality[Quality]or 1], Value);
    AddItems[Key] = nil;
  end
  Display.Data.SlotEmpty = Empty;
  Display.Data.SlotTotal = Slots;
  BaudManifestUpdateDisplayList(Display);
end


function BaudManifestCountCategorySlots(Display)
  DebugMsg("Counting slot usage for all categories.");
  for Key, Value in ipairs(Display.Data)do
    Value.SlotsUsed = nil;
  end

  local function CountCategory(Category, Path)
    Category.SlotsUsed = 0;
    Category.Path = Path;
    local Index = 1;
    while Category[Index]do
      if Category[Index].ItemString and Category[Index].Count then
        Category.SlotsUsed = Category.SlotsUsed + (Category[Index].Slots or 1);
      elseif Category[Index].Category then
        if not Display.Data[Category[Index].Category].SlotsUsed then
          Category.SlotsUsed = Category.SlotsUsed + CountCategory(Display.Data[Category[Index].Category], (Path and Path.."\\" or "")..Display.Data[Category[Index].Category].Name);
        else
          DEFAULT_CHAT_FRAME:AddMessage("BaudManifest: Removing duplicate category entry.",1,1,0);
          tremove(Category,Index);
          Index = Index - 1;
        end
      end
      Index = Index + 1;
    end
    return Category.SlotsUsed;
  end
  CountCategory(Display.Data[1]);

  local Found;
  for Category, Value in ipairs(Display.Data)do
    if not Value.SlotsUsed then
      DEFAULT_CHAT_FRAME:AddMessage("BaudManifest: Found stray category - restoring.",1,1,0);
      Found = true;
      tinsert(Display.Data[1],{Category = Category});
    end
  end
  if Found then
    return BaudManifestCountCategorySlots(Display);
  end
end


function BaudManifestCustomFilter(ItemString, Filter)
  if Filter and strfind(Filter,"%S")then
    Filter = strlower(Filter);
    Tooltip:SetOwner(WorldFrame, "ANCHOR_NONE");
    Tooltip:SetHyperlink(ItemString);
    for Index = 1, Tooltip:NumLines()do
      if strfind(strlower(getglobal("BaudManifestTooltipTextLeft"..Index):GetText()or ""),Filter,1,true)then
        return true;
      end
      if strfind(strlower(getglobal("BaudManifestTooltipTextRight"..Index):GetText()or ""),Filter,1,true)then
        return true;
      end
    end
    return false;
  end
  return true;
end


local List, Index, Table, Sort;
local function AddEntry(Key, Category, Indent)
  Index = Index + 1;
  Table = List[Index];
  if not Table then
    Table = {};
    List[Index] = Table;
  end
  Table.Index = Key;
  Table.Category = Category;
  Table.Indent = Indent;
end


function BaudManifestUpdateDisplayList(Display)
  if not Display:IsShown()then
    DebugMsg("Not updating display list while the list is closed.");
    return;
  end

  DebugMsg("Generating display list.");
  BaudManifestCountCategorySlots(Display);
  List = Display.List;
  Index = 0;
  Display.UpdateList = nil;

  local HideCategories = BaudManifestCategoriesDisabled(Display);

  local function AddCategory(Category, Indent)
    for Key, Value in ipairs(Display.Data[Category])do
      if Value.Category then
        if not HideCategories then
          AddEntry(Key, Category, Indent);
          if Display.Data[Value.Category].Expanded then
            AddCategory(Value.Category, Indent + 1);
          end
        else
          AddCategory(Value.Category,0);
        end
      elseif Value.ItemString and(Value.Count or Value.Show)and((PlayerData.HideEquipped==false)or HideCategories or(Display:GetID() > 2)or not Value.Equipped or(Value.Count > 1))and
      (BaudManifestCustomFilter(Value.ItemString, Display.CustomFilter)and((Display.Filter==1)or(Display.Filter==select(6,GetItemInfo(Value.ItemString)))))then
        AddEntry(Key, Category, Indent);
        if(Display.Sort~=1)then
          --Items that enter your backpack which the client has never seen before will not have a lot of GetItemInfo stuff, so
          --unsortable entries will not be shown
          Table.Sort = Sorts[Display.Sort].Func(Value);
        end
      end
    end
  end
  AddCategory(1,0);
  for Key = Index + 1, #List do
    List[Key] = nil;
  end

  if(Display.Sort~=1)then
    for Key, Value in ipairs(List)do
      List[Key].Key = Key;
    end
    table.sort(List,function(a,b) return(a.Sort==b.Sort)and(a.Key < b.Key)or(a.Sort < b.Sort); end);
  end
  DebugMsg("Display list finished.  Updating visual.");
  BaudManifestScrollBar_Update(Display);
end


local Feedback;
function BaudManifestScrollBar_OnVerticalScroll()
  if Feedback then
    ChatFrame1:AddMessage("Feedback warning 1");
    return;
  end
  Feedback = true;
  FauxScrollFrame_OnVerticalScroll(16, BaudManifestScrollBar_Update);
  Feedback = nil;
end


function BaudManifestScrollBar_Update(Display)
  if not Display then
    Display = this:GetParent();
  end
  if not Display:IsShown()or not Display.List then
    return;
  end
  if not Display.UpdateList then
    Display.RefreshFrame:Hide();
  end
  if(BaudManifestCategoryEditBox.Display==Display)then
    BaudManifestCategoryEditBox:Hide();
  end
  local Index, Name, Quality, MaxCount, Texture, R, G, B, Text;
  local Button, BlizItem, BlizInv, Cooldown, ButtonText, Icon;

  Text = getglobal(Display:GetName().."Text");
  Name = DisplayNames[(Display:GetID()-1)%2 + 1];
  if Display.Data.SlotEmpty then
    Name = Name.." ("..Display.Data.SlotEmpty.."/"..Display.Data.SlotTotal..")";
  end
  if(Display:GetID()==1)or(Display:GetID()==2)and BankOpen then
    Text:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b);
  else
    Name = Name.." - Offline";
    Text:SetTextColor(RED_FONT_COLOR.r,RED_FONT_COLOR.g,RED_FONT_COLOR.b);
  end
  Text:SetText(Name);
  local FrameName = Display:GetName();
  local ScrollBar = getglobal(FrameName.."ScrollBar");
  local Total = #Display.List;
  FauxScrollFrame_Update(ScrollBar, Total, Display.Entries, 16);
  local Offset = FauxScrollFrame_GetOffset(ScrollBar);
  local NoEntries = getglobal(Display:GetName().."NoEntries");
  local UpArrow = getglobal(FrameName.."ScrollBarUpArrow");
  local DownArrow = getglobal(FrameName.."ScrollBarDownArrow");

  if(Total<=0)then
    NoEntries:Show();
  else
    NoEntries:Hide();
  end

  if(Offset > 0)and(PlayerData.ShowArrows~=false)then
    UpArrow:Show();
    Display.MinEntry = 2;
  else
    UpArrow:Hide();
    Display.MinEntry = 1;
  end
  if(Offset + Display.Entries + 1 <= Total)and(PlayerData.ShowArrows~=false)then
    DownArrow:ClearAllPoints();
    DownArrow:SetPoint("BOTTOM", getglobal(FrameName.."Entry"..Display.Entries), "BOTTOM", 0, 1.6);
    DownArrow:Show();
    Display.MaxEntry = Display.Entries - 1;
  else
    DownArrow:Hide();
    Display.MaxEntry = Display.Entries;
  end

  local ItemInfo, ListItem;
  for Line = 1, Display.Entries do
    Index = Line + Offset;
    Button = getglobal(FrameName.."Entry"..Line);
    if(Index > Total)or(Line==1)and UpArrow:IsShown()or(Line==Display.Entries)and DownArrow:IsShown()then
      Button:Hide();
    else
      ListItem = Display.List[Index];
      Button.ItemIndex = ListItem.Index;
      Button.Category = ListItem.Category;
      ItemInfo = Display.Data[Button.Category][Button.ItemIndex];
      BlizItem = getglobal(Button:GetName().."BlizItem");
      BlizInv = getglobal(Button:GetName().."BlizInv");
      Icon = getglobal(Button:GetName().."Texture");
      Cooldown = getglobal(Button:GetName().."Cooldown");
      ButtonText = getglobal(Button:GetName().."Text");
      ButtonText:Show();
      Button:SetPoint("LEFT",ScrollBar,"LEFT",ListItem.Indent * IndentSize,0);
      if ItemInfo.ItemString then
        Button:SetNormalTexture("");
        Button:SetPushedTexture("");
        Button:SetHighlightTexture("");
        Name, _, Quality, _, _, _, _, MaxCount, _, Texture = GetItemInfo(ItemInfo.ItemString);
        if not Name then
          Name, Quality, MaxCount, Texture = "Unknown", 0, 1, "Interface\\Icons\\INV_Misc_QuestionMark";
          Display.Refresh = GetTime() + 1;
          Display.RefreshFrame:Show();
        end
        CooldownFrame_SetTimer(Cooldown, GetItemCooldown(ItemInfo.ItemString));
        if not ItemInfo.Slot then
          BlizItem:Hide();
          BlizInv:Hide();
        elseif ItemInfo.Bag then
          BlizInv:Hide();
          Button:SetID(ItemInfo.Bag);
          BlizItem:SetID(ItemInfo.Slot);
          BlizItem:Show();
        else
          BlizItem:Hide();
          BlizInv:SetID(ItemInfo.Slot);
          BlizInv:Show();
        end
        if(MaxCount > 1)or(ItemInfo.Count~=1)then
          Text = ItemInfo.Count or 0;
          if ItemInfo.Slots then
            Text = Text.."("..ItemInfo.Slots..")";
          end
          if PlayerData.ItemNameFirst then
            Name = Name.." "..Text;
          else
            Name = Text.." "..Name;
          end
        end
        if ItemInfo.Equipped then
          Name = "|cffffD100E|r "..Name;
        end
        ButtonText:SetText(Name);
        if ItemInfo.Count then
          R, G, B = GetItemQualityColor(Quality or 1);
          ButtonText:SetTextColor(R, G, B, 1);
        else
          ButtonText:SetTextColor(0.2,0.2,0.2,1);
        end
        if ItemInfo.Count and(not ItemInfo.Locked or(Display:GetID() > 2))and((Display:GetID()~=2)or BankOpen)then
          Icon:SetVertexColor(1,1,1);
        else
          Icon:SetVertexColor(0.5,0.5,0.5);
        end
        Icon:SetTexture(Texture);
        Icon:Show();
        if GameTooltip:IsOwned(Button)then
          BaudManifestItemEntry_OnEnter(Button, true);
        end
        Button:Show();
      else
        BlizItem:Hide();
        BlizInv:Hide();
        ButtonText:SetText(Display.Data[ItemInfo.Category].Name.." ("..Display.Data[ItemInfo.Category].SlotsUsed..")");
        ButtonText:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b,1);
        Icon:Hide();
        Cooldown:Hide();
        Text = Display.Data[ItemInfo.Category].Expanded and "Minus" or "Plus";
        Button:SetNormalTexture("Interface\\Buttons\\UI-"..Text.."Button-Up");
        Button:SetPushedTexture("Interface\\Buttons\\UI-"..Text.."Button-Down");
        Button:SetHighlightTexture("Interface\\Buttons\\UI-PlusButton-Hilight");
        Button:Show();
      end
    end
  end
end


local LastWheelValue;
local LastWheelTime = 0;
function BaudManifestScrollBar_OnMouseWheel(self, Value)
  --This is to eliminate the problem some mice have with the wheel "bouncing" back one notch when scrolling.  My mouse does this.
  --Additinaly, this function adjusts the scroll rate from what is standardly used.
  if(LastWheelValue~=Value)then
    LastWheelValue = Value;
    if(LastWheelTime + 0.1 > GetTime())then
      return;
    end
  end
  LastWheelTime = GetTime();
  local ScrollBar = getglobal(self:GetName().."ScrollBar");
  ScrollBar:SetValue(ScrollBar:GetValue() - 16 * 2 * Value);
end


function BaudManifestOptionsButton_OnClick(self)
  if not ColorPickerFrame:IsShown()then
    ToggleDropDownMenu(1, self:GetParent(), BaudManifestOptionsDropDown, self:GetName(), 0, 0);
  end
end


local ColorChangeDisplay;
function BaudManifestOptionsDropDown_Initialize()
  local Display = UIDROPDOWNMENU_MENU_VALUE;
  if not Display then
    return;
  end

  local info = UIDropDownMenu_CreateInfo(); --This function no longer causes taint problems
  info.arg1 = Display;

  if(UIDROPDOWNMENU_MENU_LEVEL==2)then
    info.func = function(Display, Texture)
      Display.Data.BGTexture = Texture;
      BaudManifestUpdateBGTexture(Display);
      CloseDropDownMenus();
    end
    --info.keepShownOnClick = true;
    for Key, Value in ipairs(Backgrounds)do
      info.text = Value.File or "None";
      info.arg2 = Value.File;
      info.checked = (Display.Data.BGTexture==Value.File);
      UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
    end
    return;
  end

  info.notCheckable = true
  if(Display:GetID()==1)and not BaudManifestDisplay2:IsShown()then
    info.text = "Show Bank";
    info.func = function()
      BaudManifestDisplay2:Show();
    end
    UIDropDownMenu_AddButton(info);
  end

  info.text = "Create Category";
  info.func = BaudManifestGetCategory;
  UIDropDownMenu_AddButton(info);

  info.text = "Categorize By Type";
  info.func = function(Display)
    if BaudManifestCategoriesDisabled(Display)then
      UIErrorsFrame:AddMessage("Cannot create categories while the list is sorted or filtered.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
      return;
    end
    local Dialog = StaticPopup_Show("BAUDMANIFEST_CATEGORIZE");
    if Dialog then
      Dialog.data = Display;
    end
  end
  UIDropDownMenu_AddButton(info);

  if(Display:GetID() <= 2)then
    info.text = (Display.Data.LockSize and "Unl" or "L").."ock Size";
    info.func = function(Display)
      Display.Data.LockSize = not Display.Data.LockSize;
    end
    UIDropDownMenu_AddButton(info);
  end

  ColorChangeDisplay = Display;
  info.text = "Color";
  info.checked = nil;
  info.hasColorSwatch = 1;
  info.r, info.g, info.b, info.opacity = unpack(Display.Data.BGColor);
--  ChatFrame1:AddMessage("Recording colors: "..(info.r or "nil")..", "..(info.g or "nil")..", "..(info.b or "nil"));
  info.opacity = 1 - info.opacity;
  info.swatchFunc = function()
    local BGColor = ColorChangeDisplay.Data.BGColor;
    BGColor[1], BGColor[2], BGColor[3] = ColorPickerFrame:GetColorRGB();
    BaudManifestUpdateBGColor(ColorChangeDisplay);
  end
  info.func = function()
    UIDropDownMenuButton_OpenColorPicker(this);
  end
  info.hasOpacity = 1;
  info.opacityFunc = function()
    ColorChangeDisplay.Data.BGColor[4] = 1 - OpacitySliderFrame:GetValue();
    BaudManifestUpdateBGColor(ColorChangeDisplay);
  end
  info.cancelFunc = function(previousValues)
    ColorChangeDisplay.Data.BGColor = {previousValues.r, previousValues.g, previousValues.b, 1 - previousValues.opacity};
    BaudManifestUpdateBGColor(ColorChangeDisplay);
  end
  UIDropDownMenu_AddButton(info);

  info = UIDropDownMenu_CreateInfo();
  info.notCheckable = true;
  info.text = "Background";
  info.hasArrow = true;
  info.value = Display;
  UIDropDownMenu_AddButton(info);
  info.hasArrow = nil;

  info.text = "Other Characters...";
  info.func = BaudManifestToggleChars;
  UIDropDownMenu_AddButton(info);

  info.text = "Options...";
  info.func = function()
    BaudManifestOptions:Show();
  end
  UIDropDownMenu_AddButton(info);
end


function BaudManifestUpdateBGColor(Display)
  Display:SetBackdropColor(unpack(Display.Data.BGColor));
  if Display.BagsFrame then
    Display.BagsFrame:SetBackdropColor(unpack(Display.Data.BGColor));
  end
end


function BaudManifestUpdateBGTexture(Display)
  local Backdrop = {
    edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
    tile = true, tileSize = 64, edgeSize = 16,
    insets = {left = 5, right = 5, top = 5, bottom = 5}
  };
  if Display.Data.BGTexture then
    Backdrop.bgFile = "Interface\\AddOns\\BaudManifest\\Backgrounds\\"..Display.Data.BGTexture..".tga";
  else
    Backdrop.bgFile = "Interface\\Buttons\\WHITE8X8";
  end
  for Key, Value in ipairs(Backgrounds)do
    if(Value.File==Display.Data.BGTexture)then
      Backdrop.tileSize = Value.Size / 2;
      break;
    end
  end
  Display:SetBackdrop(Backdrop);
  if Display.BagsFrame then
    Display.BagsFrame:SetBackdrop(Backdrop);
  end
  BaudManifestUpdateBGColor(Display);
end


function BaudManifestSortDropDown_Initialize()
  --Note that when this function runs when the addon is loaded, it determines the text to be displayed on the drop down
  local Display = getglobal(UIDROPDOWNMENU_INIT_MENU):GetParent();

  local info = UIDropDownMenu_CreateInfo();
  info.func = BaudManifestSortDropDown_OnClick;

  for Key, Value in ipairs(Sorts)do
    info.text = Value.Name;
    info.value = Key;
    info.checked = (info.value == Display.Sort)and 1 or nil;
    UIDropDownMenu_AddButton(info);
  end

  local Text = getglobal(UIDROPDOWNMENU_INIT_MENU.."Text");
end


function BaudManifestSortDropDown_OnClick()
  BaudManifestSortDropDown_Set(getglobal(UIDROPDOWNMENU_OPEN_MENU), this.value, true);
end


function BaudManifestSortDropDown_Set(DropDown, Value, Update)
  DropDown:GetParent().Sort = Value;
  if not Update then
    UIDropDownMenu_Initialize(DropDown, BaudManifestSortDropDown_Initialize);
  end
  UIDropDownMenu_SetSelectedValue(DropDown, Value);
  local Text = getglobal(DropDown:GetName().."Text");
  if(Value==1)then
    Text:SetTextColor(1,1,1);
  else
    Text:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b);
  end
  if Update then
    BaudManifestUpdateDisplayList(DropDown:GetParent());
    BaudManifestResetButton_Update(DropDown:GetParent());
  end
end


function BaudManifestFilterDropDown_Initialize()
  --Note that when this function runs when the addon is loaded, it determines the text to be displayed on the drop down
  local Display = getglobal(UIDROPDOWNMENU_INIT_MENU):GetParent();

  local info = UIDropDownMenu_CreateInfo();
  info.func = BaudManifestFilterDropDown_OnClick;

  info.text = "No Filter";
  info.value = 1;
  info.checked = (info.value == Display.Filter)and 1 or nil;
  UIDropDownMenu_AddButton(info);

  if not Display.Data then
    return;
  end

  local Text = getglobal(UIDROPDOWNMENU_INIT_MENU.."Text");
  if(Display.Filter~=1)then
    Text:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b);
  else
    Text:SetTextColor(1,1,1);
  end

  local ItemTypes = {};
  local Type;
  for Category, Value in ipairs(Display.Data)do
    for ItemIndex, Value in ipairs(Value)do
      if Value.ItemString then
        Type = select(6,GetItemInfo(Value.ItemString));
        for Key, Value in ipairs(ItemTypes)do
          if(Type==Value)then
            Type = nil;
            break;
          end
        end
        if Type then
          tinsert(ItemTypes,Type);
        end
      end
    end
  end

  table.sort(ItemTypes,function(a,b) return (a < b);end);

  for Key, Value in ipairs(ItemTypes)do
    info.text = Value;
    info.value = Value;
    info.checked = (info.value == Display.Filter)and 1 or nil;
    UIDropDownMenu_AddButton(info);
  end
end


function BaudManifestFilterDropDown_OnClick()
  BaudManifestFilterDropDown_Set(getglobal(UIDROPDOWNMENU_OPEN_MENU), this.value, true);
end


function BaudManifestFilterDropDown_Set(DropDown, Value, Update)
  DropDown:GetParent().Filter = Value
  if not Update then
    UIDropDownMenu_Initialize(DropDown, BaudManifestFilterDropDown_Initialize);
  end
  UIDropDownMenu_SetSelectedValue(DropDown, Value);
  local Text = getglobal(DropDown:GetName().."Text");
  if(Value==1)then
    Text:SetTextColor(1,1,1);
  else
    Text:SetTextColor(NORMAL_FONT_COLOR.r,NORMAL_FONT_COLOR.g,NORMAL_FONT_COLOR.b);
  end
  if Update then
    BaudManifestUpdateDisplayList(DropDown:GetParent());
    BaudManifestResetButton_Update(DropDown:GetParent());
  end
end


function BaudManifestFilterEditBox_OnTextChanged(self)
  local Display = self:GetParent();
  if((Display.CustomFilter or "")==self:GetText())then
    return;
  end
  Display.CustomFilter = self:GetText();
  Display.Refresh = GetTime() + FilterDelay;
  Display.RefreshFrame:Show();
  Display.UpdateList = true;
  BaudManifestResetButton_Update(Display);
end


function BaudManifestResetButton_OnClick(Display)
  PlaySound("igMainMenuOptionCheckBoxOff");
  getglobal(Display:GetName().."FilterEditBox"):SetText("");
  BaudManifestFilterDropDown_Set(getglobal(Display:GetName().."FilterDropDown"), 1);
  BaudManifestSortDropDown_Set(getglobal(Display:GetName().."SortDropDown"), 1)
  BaudManifestClearFocus();
  BaudManifestUpdateDisplayList(Display);
  BaudManifestResetButton_Update(Display);
end


function BaudManifestResetButton_Update(Display)
  local Button = getglobal(Display:GetName().."ResetButton");
  if BaudManifestCategoriesDisabled(Display)then
    Button:Enable();
  else
    Button:Disable();
  end
end


function BaudManifestCharacters_OnLoad(self)
  self:RegisterForDrag("LeftButton");
  self:RegisterForClicks("RightButtonUp", "LeftButtonUp");
  tinsert(UISpecialFrames, self:GetName()); --UIMenus
end


function BaudManifestGenericClick(Button)
  if(Button=="RightButton")then
    BaudManifestClearVirtualDrag();
  end
  BaudManifestClearFocus();
end


function BaudManifestCharacters_OnShow()
--  BaudManifestCharactersScrollBoxHighlight:Hide();
end


function BaudManifestCharacterEntry_OnClick(self)
  BaudManifestClearVirtualDrag();
  SelectedChar = self.Name;
  local Data = GetCharData(SelectedChar);
  if(SelectedChar=="All Characters")then
    if not AllCharData then
      AllCharData = Data;
    end
    BaudManifestUpdateAllChars();
  end
  BaudManifestCharactersScrollBoxHighlight:SetPoint("TOP",self);
  BaudManifestCharactersScrollBoxHighlight:Show();
  local Display;
  for Index = 1, 2 do
    Display = getglobal("BaudManifestDisplay"..(Index + 2));
    if(Index~=1)and(SelectedChar=="All Characters")then
      Display:Hide();
      break;
    end
    getglobal(Display:GetName().."CloseButton"):Hide();
    getglobal(Display:GetName().."OptionsButton"):SetPoint("TOPRIGHT",-5,-5);
    Display.Name = SelectedChar;
    Display.Data = Data[Index];
    BaudManifestClearZeros(Display);
    BaudManifestUpdateBGTexture(Display);
    BaudManifestUpdateDisplayList(Display);
    if(Index==1)then
      SetMoneyFrameColor(Display:GetName().."MoneyFrame", 1.0, 1.0, 1.0);
      MoneyFrame_Update(Display:GetName().."MoneyFrame", (Data.Money or 0));
    end
    Display:Show();
  end
end


function BaudManifestUpdateAllChars()
  DebugMsg("Updating combined character list");
  local ItemData, AllChars;
  if not AllCharHash then
    DebugMsg("Creating Hashd tables");
    AllCharHash = {};
    OtherCharMoney = 0;

    for Category, Value in ipairs(AllCharData[1])do
      for ItemIndex, Value in ipairs(Value)do
        if Value.ItemString then
          AllCharHash[Value.ItemString] = Value;
          Value.OtherChars = nil;
        end
      end
    end

    for Key, Value in pairs(RealmData)do
      if(Value~=AllCharData)and(Value~=PlayerData)then
        OtherCharMoney = OtherCharMoney + (Value.Money or 0);
        for Display, Value in ipairs(Value)do
          for Category, Value in ipairs(Value)do
            for Key, Value in ipairs(Value)do
              if Value.ItemString and Value.Count and(Value.Count > 0)then
                ItemData = AllCharHash[Value.ItemString];
                if ItemData then
                  ItemData.OtherChars = Value.Count + (ItemData.OtherChars or 0);
                else
                  ItemData = {ItemString = Value.ItemString, SuffixFactor = Value.SuffixFactor, OtherChars = Value.Count};
                  tinsert(AllCharData[1][1], ItemData);
                  AllCharHash[Value.ItemString] = ItemData;
                end
              end
            end
          end
        end
      end
    end
  end

  for Key, Value in pairs(AllCharHash)do
    Value.Count = Value.OtherChars;
  end

  AllCharData.Money = OtherCharMoney + PlayerData.Money;

  for Display, Value in ipairs(PlayerData)do
    for Category, Value in ipairs(Value)do
      for Key, Value in ipairs(Value)do
        if Value.ItemString and Value.Count and(Value.Count > 0)then
          ItemData = AllCharHash[Value.ItemString];
          if ItemData then
            ItemData.Count = (ItemData.Count or 0) + Value.Count;
          else
            ItemData = {ItemString = Value.ItemString, Count = Value.Count, SuffixFactor = Value.SuffixFactor};
            tinsert(AllCharData[1][1], ItemData);
            AllCharHash[Value.ItemString] = ItemData;
          end
        end
      end
    end
  end

  local Root = AllCharData[1][1];
  local Index = 0;
  while true do
    Index = Index + 1;
    if not Root[Index]then
      break;
    end
    if Root[Index].ItemString and not Root[Index].Count then
      AllCharHash[Root[Index].ItemString] = nil;
      tremove(Root, Index);
      Index = Index - 1;
    end
  end
end


function BaudManifestOptions_OnLoad(self)
  tinsert(UISpecialFrames, self:GetName()); --UIMenus

  self:RegisterForDrag("LeftButton");
  self:RegisterForClicks("RightButtonUp", "LeftButtonUp");

  local Text, DropDown;
  local QText = {[0] = "grey", [1] = "white", [2] = "green"}
  for Quality = 0, 2 do
    DropDown = CreateFrame("Frame","BaudManifestQualityDropDown"..Quality,self,"UIDropDownMenuTemplate");
    DropDown:SetPoint("TOPLEFT",155,-35-Quality*30);
    DropDown:SetID(Quality);
    UIDropDownMenu_SetWidth(110, DropDown);
    Text = DropDown:CreateFontString(DropDown:GetName().."QualityText"..Quality,"ARTWORK","GameFontHighlight");
    Text:SetText("Put new "..select(4,GetItemQualityColor(Quality))..QText[Quality].."|r items in");
    Text:SetPoint("RIGHT",DropDown,"LEFT",10,0);
  end

  DropDown = CreateFrame("Frame","BaudManifestAutoSellDropDown",self,"UIDropDownMenuTemplate");
  DropDown:SetPoint("TOPLEFT", 155, -140);
  UIDropDownMenu_SetWidth(110, DropDown);
  Text = DropDown:CreateFontString(DropDown:GetName().."Description", "ARTWORK", "GameFontHighlight");
  Text:SetText("Sell items in");
  Text:SetPoint("RIGHT",DropDown,"LEFT",10,0);

  local Button;
  for Index, Value in ipairs(CheckButtons)do
    Button = CreateFrame("CheckButton",self:GetName().."CheckButton"..Index,self,"OptionsCheckButtonTemplate");
    Button:SetPoint("TOPRIGHT",self,-160+(Value.Depend and 10 or 0),-25*Index);
    Button:SetID(Index);
    Button.tooltipText = Value.TooltipText;
    getglobal(Button:GetName().."Text"):SetText(Value.Text);
    Button:SetScript("OnClick",BaudManifestCheckButton_OnClick);
  end
  self:SetHeight(70 + #CheckButtons * 25);
end


local BankBlocked = false;
function BaudManifestUpdateBlizBankBlock()
  local Block = PlayerData.ShowBank and PlayerData.HideBlizBank;
  if(BankBlocked==Block)then
    return;
  end
  DebugMsg("Bliz bank blocked: "..(Block and "On" or "Off"));
  BankBlocked = Block;
  if Block then
    BankFrame:UnregisterEvent("BANKFRAME_OPENED");
  else
    BankFrame:RegisterEvent("BANKFRAME_OPENED");
  end
end


function BaudManifestOptions_OnShow(self)
  self:ClearAllPoints();
  self:SetPoint("CENTER");
  self:Raise();

  local DropDown;
  for Quality = 0, 2 do
    DropDown = getglobal("BaudManifestQualityDropDown"..Quality);
    UIDropDownMenu_Initialize(DropDown, BaudManifestQualityDropDown_Initialize);
    UIDropDownMenu_SetSelectedValue(DropDown, PlayerData.PutQuality[Quality]);
  end
  DropDown = BaudManifestAutoSellDropDown;
  UIDropDownMenu_Initialize(DropDown, BaudManifestAutoSellDropDown_Initialize);
  UIDropDownMenu_SetSelectedValue(DropDown, PlayerData.AutoSell);
  local Button;
  for Index, Value in ipairs(CheckButtons)do
    Button = getglobal(self:GetName().."CheckButton"..Index);
    Button:SetChecked(PlayerData[Value.SavedVar]);
  end
  BaudManifestOptionsUpdateDepend();

  BaudManifestOptionsScaleSliderLow:SetText("50%");
  BaudManifestOptionsScaleSliderHigh:SetText("200%");
  BaudManifestOptionsScaleSlider.tooltipText = "Scale of the manifest.";
  BaudManifestOptionsScaleSlider:SetValue(PlayerData.Scale);
end


function BaudManifestOptionsUpdateDepend()
  for Index, Value in ipairs(CheckButtons)do
    if Value.Depend then
      Button = getglobal("BaudManifestOptionsCheckButton"..Index);
      if PlayerData[Value.Depend]then
        OptionsFrame_EnableCheckBox(Button);
      else
        OptionsFrame_DisableCheckBox(Button);
      end
    end
  end
end


local CategoryList = {};
function BaudManifestMakeCategoryDropDown(FirstItem, Selected, Func)
  for Key, Value in ipairs(CategoryList)do
    CategoryList[Key] = nil;
  end

  local info = UIDropDownMenu_CreateInfo();
  info.func = Func;

  for Category, Value in ipairs(PlayerData[1])do
    if(Category > 1)then
      tinsert(CategoryList,{Name=Value.Name,Category=Category});
    end
  end
  table.sort(CategoryList, function(a,b)return(a.Name < b.Name)end);
  tinsert(CategoryList, 1, FirstItem);

  for Index, Value in ipairs(CategoryList)do
    info.text = Value.Name;
    info.value = Value.Category;
    info.checked = (info.value == Selected)and 1 or nil;
    UIDropDownMenu_AddButton(info);
  end
end


function BaudManifestQualityDropDown_Initialize()
  local Quality = getglobal(UIDROPDOWNMENU_INIT_MENU):GetID();
  BaudManifestMakeCategoryDropDown({Name="<Inventory>",Category=1}, PlayerData.PutQuality[Quality], BaudManifestQualityDropDown_OnClick);
end


function BaudManifestQualityDropDown_OnClick()
  local DropDown = getglobal(UIDROPDOWNMENU_OPEN_MENU);
  PlayerData.PutQuality[DropDown:GetID()] = this.value;
  UIDropDownMenu_SetSelectedValue(DropDown, this.value);
end


function BaudManifestAutoSellDropDown_Initialize()
  BaudManifestMakeCategoryDropDown({Name="<Disabled>",Category=1}, PlayerData.AutoSell, BaudManifestAutoSellDropDown_OnClick);
end


function BaudManifestAutoSellDropDown_OnClick()
  PlayerData.AutoSell = this.value;
  UIDropDownMenu_SetSelectedValue(getglobal(UIDROPDOWNMENU_OPEN_MENU), PlayerData.AutoSell);
end


local TooltipMoney;
function BaudManifestTooltip_OnTooltipAddMoney(Money)
  TooltipMoney = Money;
end


function BaudManifestToMoney(Money)
  Money = floor(Money);
  local Text = "";
  if(Money % 100 > 0)and(Money < 10000)or(Money==0)then
    --203,163,133
    Text = "|cffba795a"..(Money % 100).."c|r";
  end
  Money = floor(Money / 100);
  if(Money % 100 > 0)and(Money < 100000)then
    --C0C0C0
    Text = (Money % 100).."s"..Text;
  end
  Money = floor(Money / 100);
  if(Money > 0)then
    --227,216,128
    Text = "|cfff0f020"..Money.."g|r"..Text;
  end
  return Text;
end


function BaudManifestAutoSellItems()
  if(PlayerData.AutoSell <= 1)then
    return;
  end
  local AutoSell = PlayerData[1][PlayerData.AutoSell];
  local Total = 0;
  local Profit = 0;
  local Item;
  BaudManifestForEachSlot(BaudManifestDisplay1,
    function(Bag, Slot)
      Item = GetContainerItemLink(Bag, Slot);
      if not Item then
        return;
      end
      Item = strmatch(Item, ItemStringMatch);
      if not Item then
        return;
      end
      for Key, Value in ipairs(AutoSell)do
        if(Value.ItemString==Item)then
          TooltipMoney = nil;
          Tooltip:SetOwner(WorldFrame, "ANCHOR_NONE");
          Tooltip:SetBagItem(Bag, Slot);
          if TooltipMoney then
            Profit = Profit + TooltipMoney;
            UseContainerItem(Bag, Slot);
            Total = Total + 1;
          end
          return;
        end
      end
    end,
    function(Bag)
      return (not Bag);
    end
  );
  if(Total > 0)then
    DEFAULT_CHAT_FRAME:AddMessage(MsgPrefix.."Selling "..Total.." item"..((Total==1)and "" or "s").." for "..BaudManifestToMoney(Profit)..".");
  end
end


function BaudManifestCheckButton_OnClick(self)
  if self:GetChecked()then
    PlaySound("igMainMenuOptionCheckBoxOff");
  else
    PlaySound("igMainMenuOptionCheckBoxOn");
  end
  local SavedVar = CheckButtons[self:GetID()].SavedVar;
  PlayerData[SavedVar] = self:GetChecked()and true or false;
  if(SavedVar=="HideEquipped")then
    BaudManifestUpdateDisplayList(BaudManifestDisplay1);
  elseif(SavedVar=="ItemNameFirst")or(SavedVar=="ShowArrows")then
    for Display = 1, 4 do
      BaudManifestScrollBar_Update(getglobal("BaudManifestDisplay"..Display));
    end
  elseif(SavedVar=="ReplaceBlizBags")and PlayerData[SavedVar]then
    BaudManifestHookBlizBags();
  end
  BaudManifestUpdateBlizBankBlock();
  BaudManifestOptionsUpdateDepend();
end


function BaudManifestUpdateScales()
  local Display, X, Y, Multiplier;
  for Index = 1, 4 do
    Display = getglobal("BaudManifestDisplay"..Index);
    Multiplier = Display:GetScale() / (PlayerData.Scale / 100);
    X, Y = Display:GetCenter();
    Display:SetScale(PlayerData.Scale / 100);
    if(Index <= 2)then
      Display:ClearAllPoints();
      Display:SetPoint("CENTER",UIParent,"BOTTOMLEFT", X * Multiplier, Y * Multiplier);
    end
    BaudManifestDisplay_OnSizeChanged(Display);
  end
end


function BaudManifestScaleSlider_OnValueChanged(self)
  getglobal(self:GetName().."Text"):SetText("Scale: "..self:GetValue().."%");
  if IgnoreHooks then
    return;
  end
  PlayerData.Scale = self:GetValue();
  BaudManifestUpdateScales();
end


--if enabled, this will automatically hide blizzard's backpack when it pops up at the mail or vendor.  The actual opening is
--not prevented to avoid the posibility of spreading taint.
hooksecurefunc("OpenBackpack",function()
  if not BlizBagsHooked and PlayerData.HideBlizInv and((this==MailFrame)or(this==MerchantFrame))and not ContainerFrame1.backpackWasOpen then
    for Index = 1, NUM_CONTAINER_FRAMES do
      local Frame = getglobal("ContainerFrame"..Index);
      if Frame:IsShown()and(Frame:GetID()==0)then
        Frame:Hide();
      end
    end
  end
end);

--/script BaudManifestEmptyBag(BaudManifestDisplay1, 4)
function BaudManifestEmptyBag(Display, BagToEmpty, Unequip)
  DebugMsg("Emptying bag "..BagToEmpty);
  BaudManifestReadSpecials(Display, BagToEmpty);

  local Link, BestBag, ItemID, InvID;
  local MaxSlots = GetContainerNumSlots(BagToEmpty) + 1;
  local NextSlot = 1;
  local NotEmpty;
  if not BaudManifestForEachSlot(Display,
    function(Bag, Slot)
      Link = GetContainerItemLink(Bag, Slot);
      if Link then
        return;
      end
      for Slot2 = NextSlot, MaxSlots do
        if(Slot2>=MaxSlots)then
          if not Unequip then
            return true;
          end
          Link = GetInventoryItemLink("player", ContainerIDToInventoryID(BagToEmpty));
        else
          Link = GetContainerItemLink(BagToEmpty, Slot2);
          if Link then
            NotEmpty = true;
            if select(3, GetContainerItemInfo(BagToEmpty, Slot2))then
              Link = nil;
            end
          end
        end
        if Link then
          NextSlot = Slot2 + 1;
          ItemID = tonumber(strmatch(Link, "item:(%d+):")or 0);
          if(Slot2 < MaxSlots)then
            Moved = true;
            ClearCursor();
            IgnoreHooks = true;
            PickupContainerItem(BagToEmpty, Slot2);
            IgnoreHooks = false;
          elseif NotEmpty then
            DebugMsg("Queueing bag unequip.");
            DragPending = DragItem;
            DragItem = nil;
            return true;
          else
            DebugMsg("Unequipping bag.");
            if not CursorHasItem()then
              IgnoreHooks = true;
              PickupBagFromSlot(ContainerIDToInventoryID(BagToEmpty));
              IgnoreHooks = false;
            end
          end
          if CursorHasItem()then
            for Index = 1, NumSpecials do
              if BagTypes[SpecialBags[Index].Type][ItemID]and(SpecialBags[Index].Free > 0)then
                SpecialBags[Index].Free = SpecialBags[Index].Free - 1;
                InvID = ContainerIDToInventoryID(SpecialBags[Index].Index);
                if CursorCanGoInSlot(InvID)then
                  PickupContainerItem(SpecialBags[Index].Index, SpecialBags[Index].Empty[1]);
                else
                  PutItemInBag(InvID);
                end
                tremove(SpecialBags[Index].Empty, 1);
                ItemID = nil;
                break;
              end
            end
            if ItemID then
              if(Bag==0)then
                PutItemInBackpack();
              else
                InvID = (Bag > 0)and ContainerIDToInventoryID(Bag);
                if CursorCanGoInSlot(InvID)then
                  PickupContainerItem(Bag, Slot);
                else
                  PutItemInBag(InvID);
                end
              end
              return;
            end
          end
        end
      end
      return true;
    end,
    function(Bag)
      if not Bag or(Bag < -1)or(Bag==BagToEmpty)then
        return true;
      end
      for Index = 1, NumSpecials do
        if(SpecialBags[Index].Index==Bag)then
          return true;
        end
      end
    end
  )then
    UIErrorsFrame:AddMessage("Not enough bag space to unequip that.", 1.0, 0.1, 0.1, 1.0, UIERRORS_HOLD_TIME);
    ClearCursor();
    DragItem = nil;
  else
    DebugMsg("Bag Emptying Completed");
  end
end


function BaudManifestBagsButton_OnClick(self)
  local Display = self:GetParent();
  Display.Data.ShowBags = self:GetChecked();
  if Display.Data.ShowBags then
    Display.BagsFrame:Show();
  else
    Display.BagsFrame:Hide();
  end
end


function BaudManifestBankPurchaseUpdate()
  DebugMsg("Updating bank purchase frame.");

  local Purchase = BaudManifestBankSlotPurchaseButton;
  local BagsFrame = Purchase:GetParent();
  local Slots, Full = GetNumBankSlots();

  local BagSlot;

  for Bag = 1, NUM_BANKBAGSLOTS do
    BagSlot = getglobal("BaudMBankBag"..Bag);

    if(Bag <= Slots)then
      SetItemButtonTextureVertexColor(BagSlot, 1.0, 1.0, 1.0);
      BagSlot.tooltipText = BANK_BAG;
    else
      SetItemButtonTextureVertexColor(BagSlot, 1.0, 0.1, 0.1);
      BagSlot.tooltipText = BANK_BAG_PURCHASE;
    end
  end

  if Full then
    Purchase:Hide();
    BagsFrame:SetHeight(BagsFrame.Height);
    return;
  end

  local Cost = GetBankSlotCost(Slots);

  --This line allows the confirmation box to show the cost
  BankFrame.nextSlotCost = Cost;

  local MoneyFrame = Purchase:GetName().."MoneyFrame";
  if(GetMoney() >= Cost)then
    SetMoneyFrameColor(MoneyFrame, 1.0, 1.0, 1.0);
  else
    SetMoneyFrameColor(MoneyFrame, 1.0, 0.1, 0.1);
  end
  MoneyFrame_Update(MoneyFrame, Cost);

  Purchase:Show();
  BagsFrame:SetHeight(BagsFrame.Height + 40);
end


local BlizBagsHooked, OpenAllBypass;
local pre_ToggleBag, pre_OpenAllBags, pre_OpenAllBags2, pre_BagSlotButton_OnClick, pre_ToggleBackpack, pre_ToggleKeyRing;
function BaudManifestHookBlizBags()
  if BlizBagsHooked then
    return;
  end
  BlizBagsHooked = true;

  local function BypassHook()
    if not PlayerData.ReplaceBlizBags or this and(strsub(this:GetName(),1,4)=="Baud")then
      return true;
    end
    BaudManifestToggleDisplay(1);
  end

  pre_ToggleBag = ToggleBag;
  ToggleBag = function(...)
    if OpenAllBypass or(select(1, ...) > 4)or this and(strsub(this:GetName(),-11)=="CloseButton")or BypassHook()then
      return pre_ToggleBag(...);
    end
  end

  --This is wierd coding, but I'm avoiding making a table to store any potential return values.
  pre_OpenAllBags = OpenAllBags;
  pre_OpenAllBags2 = function(...)
    OpenAllBypass = false;
    return ...;
  end
  OpenAllBags = function(...)
    OpenAllBypass = true;
    return pre_OpenAllBags2(pre_OpenAllBags(...));
  end

  pre_ToggleBackpack = ToggleBackpack;
  ToggleBackpack = function(...)
    --There is an alternate function to block the Blizzard bags at the mail / vendor, but it is just a taint free alternative
    if OpenAllBypass or((this~=MailFrame)and(this~=MerchantFrame)or not PlayerData.HideBlizInv)and BypassHook()then
      return pre_ToggleBackpack(...);
    end
  end

  pre_ToggleKeyRing = ToggleKeyRing;
  ToggleKeyRing = function(...)
    if BypassHook()then
      return pre_ToggleKeyRing(...);
    end
  end
end


hooksecurefunc("BankFrameItemButton_OnEnter", function(self)
  if not self.isBag or BankOpen or GameTooltip:GetItem()then
    return;
  end
  local Bag = PlayerData.BankBag[self:GetID()-4].Link;
  if not Bag then
    return;
  end
  Bag = strmatch(Bag, "(item[%d:%-]+)");
  if not Bag then
    return;
  end
  if(self:GetRight() >= (GetScreenWidth() / 2))then
    GameTooltip:SetOwner(self, "ANCHOR_LEFT");
  else
    GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
  end
  GameTooltip:SetHyperlink(Bag);
end);


function BaudManifestResetPositons()
  BaudManifestDisplay1:ClearAllPoints();
  BaudManifestDisplay1:SetPoint("LEFT", UIParent, "CENTER", 10, 0);
  BaudManifestDisplay2:ClearAllPoints();
  BaudManifestDisplay2:SetPoint("RIGHT", UIParent, "CENTER", -10, 0);
  DEFAULT_CHAT_FRAME:AddMessage(MsgPrefix.."Manifest positions reset.");
end